"""
Helper functions for pyferret stats external functions.
"""

from __future__ import print_function

import math
import numpy
try:
    import scipy.stats
    import scipy.special
except ImportError:
    # do not raise an error until actually used
    pass
import pyferret


def getdistname(distribname=None):
    """
    Translates probability distribution names into scipy.stats names.
    Arguments:
        distribname - the distribution name, or None.
    Returns:
        If distribname is given (and not None), the "scipy.stats" name for
            the probability distribution given in distribname, or None if
            the probability distribution name is not recognized or supported.
        If distribname is not given (or None), returns a list of string
            tuples, where the first name in each tuple is the scipy.stats
            name, the second name is a "full name" and any other names are
            other recognized aliases.
    """
    namelist = (
                 ( "beta", "Beta", ),
                 ( "binom", "Binomial", ),
                 ( "cauchy", "Cauchy", ),
                 ( "chi", "Chi", ),
                 ( "chi2", "Chi-Square", ),
                 ( "expon", "Exponential", ),
                 ( "exponweib", "Exponentiated-Weibull", ),
                 ( "f", "F", "Fisher", ),
                 ( "gamma", "Gamma", ),
                 ( "geom", "Geometric", "Shifted-Geometric", ),
                 ( "hypergeom", "Hypergeometric", ),
                 ( "invgamma", "Inverse-Gamma", ),
                 ( "laplace", "Laplace", ),
                 ( "lognorm", "Log-Normal", ),
                 ( "nbinom", "Negative-Binomial", ),
                 ( "norm", "Normal", ),
                 ( "pareto", "Pareto", ),
                 ( "poisson", "Poisson", ),
                 ( "randint", "Random-Integer", "Discrete-Uniform", ),
                 ( "t", "Students-T", ),
                 ( "uniform", "Uniform", ),
                 ( "weibull_min", "Weibull", ),
               )
    if distribname is None:
        return namelist
    lcdistname = str(distribname).lower()
    # Testing below verifies the above names are all recognized in the following
    if lcdistname == "beta":
        return "beta"
    if (lcdistname == "binom") or (lcdistname == "binomial"):
        return "binom"
    if lcdistname == "cauchy":
        return "cauchy"
    if lcdistname == "chi":
        return "chi"
    if (lcdistname == "chi2") or (lcdistname == "chi-square"):
        return "chi2"
    if (lcdistname == "expon") or (lcdistname == "exponential"):
        return "expon"
    if (lcdistname == "exponweib") or (lcdistname == "exponentiated-weibull"):
        return "exponweib"
    if (lcdistname == "f") or (lcdistname == "fisher"):
        return "f"
    if lcdistname == "gamma":
        return "gamma"
    if (lcdistname == "geom") or (lcdistname == "geometric") or (lcdistname == "shifted-geometric"):
        return "geom"
    if (lcdistname == "hypergeom") or (lcdistname == "hypergeometric"):
        return "hypergeom"
    if (lcdistname == "invgamma") or (lcdistname == "inverse-gamma"):
        return "invgamma"
    if lcdistname == "laplace":
        return "laplace"
    if (lcdistname == "lognorm") or (lcdistname == "log-normal"):
        return "lognorm"
    if (lcdistname == "nbinom") or (lcdistname == "negative-binomial"):
        return "nbinom"
    if (lcdistname == "norm") or (lcdistname == "normal"):
        return "norm"
    if lcdistname == "pareto":
        return "pareto"
    if lcdistname == "poisson":
        return "poisson"
    if (lcdistname == "randint") or (lcdistname == "random-integer") or (lcdistname == "discrete-uniform"):
        return "randint"
    if (lcdistname == "t") or (lcdistname == "students-t"):
        return "t"
    if lcdistname == "uniform":
        return "uniform"
    if (lcdistname == "weibull_min") or (lcdistname == "weibull"):
        return "weibull_min"
    return None


def getdistparams(distname, params, tostd=False):
    """
    Converts between "standard" parameters for a probability function and
    the tuple of numeric parameters expected for the scipy.stats function.

    Arguments:
        distname - the "scipy.stats" name for the probability distribution
        params - the parameters for the probability distribution, or None
        tostd - convert from scipy.stats to standard parameters?

    Returns:
        If params is None, a list of string pairs describing the standard
        parameters for the distribution.

        If tostd is False, a tuple of scipy.stats parameters corresponding
        to the standard parameters given in params.  The given parameters
        are explicitly converted to the correct type (thus strings could
        be given) and error checking is performed.  Additional offset and
        scaling parameters, if appropriate, can be given.

        If tostd is True, a tuple of standard parameters corresponding
        to the scipy.stats parameters given in params.  No explicit type
        conversion or error checking is performed.  Offset and scaling
        parameters, if appropriate, are assumed to be given.

    Raises:
        ValueError if the parameters are inappropriate for the probabiity
        distribution.
    """
    if distname == "beta":
        if params is None:
            return ( ( "ALPHA", "first shape", ),
                     ( "BETA", "second shape", ), )
        if tostd:
            return ( params[0], params[1], params[2], params[3], )
        if (len(params) < 2) or (len(params) > 4):
            raise ValueError("Two to four parameters expected for the Beta distribution")
        alpha = float(params[0])
        beta = float(params[1])
        if (alpha <= 0.0) or (beta <= 0.0):
            raise ValueError("Invalid parameter(s) for the Beta distribution")
        offset = 0.0
        scaling = 1.0
        try:
            offset = float(params[2])
            scaling = float(params[3])
        except IndexError:
            pass
        return ( alpha, beta, offset, scaling, )
    if distname == "binom":
        if params is None:
            return ( ( "N", "number of trials", ),
                     ( "P", "success probability in each trial", ), )
        if tostd:
            return ( params[0], params[1], )
        if len(params) != 2:
            raise ValueError("Two parameters expected for the Binomial distribution")
        nflt = float(params[0])
        prob = float(params[1])
        if (nflt < 0.0) or (prob < 0.0) or (prob > 1.0):
            raise ValueError("Invalid parameter(s) for the Binomial distribution")
        return ( nflt, prob, )
    if distname == "cauchy":
        if params is None:
            return ( ( "M", "location (median)", ),
                     ( "GAMMA", "scale (half-width at half-maximum)", ), )
        if tostd:
            return ( params[0], params[1], )
        if len(params) != 2:
            raise ValueError("Two parameters expected for the Cauchy distribution")
        m = float(params[0])
        gamma = float(params[1])
        if gamma <= 0.0:
            raise ValueError("Invalid parameter for the Cauchy distribution")
        return ( m, gamma, )
    if distname == "chi":
        if params is None:
            return ( ( "DF", "degrees of freedom", ), )
        if tostd:
            return ( params[0], params[1], params[2], )
        if (len(params) < 1) or (len(params) > 3):
            raise ValueError("One to three parameters expected for the Chi distribution")
        degfree = float(params[0])
        if degfree <= 0.0:
            raise ValueError("Invalid parameter for the Chi distribution")
        offset = 0.0
        scaling = 1.0
        try:
            offset = float(params[1])
            scaling = float(params[2])
        except IndexError:
            pass
        return ( degfree, offset, scaling, )
    if distname == "chi2":
        if params is None:
            return ( ( "DF", "degrees of freedom", ), )
        if tostd:
            return ( params[0], params[1], params[2], )
        if (len(params) < 1) or (len(params) > 3):
            raise ValueError("One to three parameters expected for the Chi-Square distribution")
        degfree = float(params[0])
        if degfree <= 0.0:
            raise ValueError("Invalid parameter for the Chi-Square distribution")
        offset = 0.0
        scaling = 1.0
        try:
            offset = float(params[1])
            scaling = float(params[2])
        except IndexError:
            pass
        return ( degfree, offset, scaling, )
    if distname == "expon":
        if params is None:
            return ( ( "LAMBDA", "rate (inverse scale)", ), )
        if tostd:
            return ( 1.0 / params[1], params[0], )
        if (len(params) < 1) or (len(params) > 2):
            raise ValueError("One or two parameters expected for the Exponential distribution")
        lambdaflt = float(params[0])
        if lambdaflt <= 0.0:
            raise ValueError("Invalid parameter for the Exponential distribution")
        try:
            offset = float(params[1])
        except IndexError:
            offset = 0.0
        return ( offset, 1.0 / lambdaflt, )
    if distname == "exponweib":
        if params is None:
            return ( ( "K", "Weibull shape", ),
                     ( "LAMBDA", "scale", ),
                     ( "ALPHA", "power shape", ), )
        if tostd:
            return ( params[1], params[3], params[0], params[2], )
        if (len(params) < 3) or (len(params) > 4):
            raise ValueError("Three or four parameters expected for the Exponentiated-Weibull distribution")
        k =  float(params[0])
        lambdaflt = float(params[1])
        alpha = float(params[2])
        if (k <= 0.0) or (lambdaflt <= 0.0) or (alpha <= 0):
            raise ValueError("Invalid parameter(s) for the Exponentiated-Weibull distribution")
        try:
            offset = float(params[3])
        except IndexError:
            offset = 0.0
        return ( alpha, k, offset, lambdaflt, )
    if distname == "f":
        if params is None:
            return ( ( "DFN", "numerator degrees of freedom", ),
                     ( "DFD", "denominator degrees of freedom", ), )
        if tostd:
            return ( params[0], params[1], params[2], params[3], )
        if (len(params) < 2) or (len(params) > 4):
            raise ValueError("Two to four parameters expected for the F distribution")
        dfnum = float(params[0])
        dfdenom = float(params[1])
        if (dfnum <= 0.0) or (dfdenom <= 0.0):
           raise ValueError("Invalid parameter(s) for the F distribution")
        offset = 0.0
        scaling = 1.0
        try:
            offset = float(params[2])
            scaling = float(params[3])
        except IndexError:
            pass
        return ( dfnum, dfdenom, offset, scaling, )
    if distname == "gamma":
        if params is None:
            return ( ( "ALPHA", "shape", ),
                     ( "THETA", "scale", ), )
        if tostd:
            return ( params[0], params[2], params[1], )
        if (len(params) < 2) or (len(params) > 3):
            raise ValueError("Two or three parameters expected for the Gamma distribution")
        alpha = float(params[0])
        theta = float(params[1])
        if (alpha <= 0.0) or (theta <= 0.0):
            raise ValueError("Invalid parameter(s) for the Gamma distribution")
        try:
            offset = float(params[2])
        except IndexError:
            offset = 0.0
        return ( alpha, offset, theta, )
    if distname == "geom":
        if params is None:
            return ( ( "P", "success probability", ), )
        if tostd:
            return ( params[0], )
        if len(params) != 1:
            raise ValueError("One parameter expected for the Shifted-Geometric distribution")
        prob = float(params[0])
        if (prob < 0.0) or (prob > 1.0):
            raise ValueError("Invalid parameter for the Shifted-Geometric distribution")
        return ( prob, )
    if distname == "hypergeom":
        if params is None:
            return ( ( "NTOTAL", "total number of items", ),
                     ( "NGOOD", "total number of 'success' items", ),
                     ( "NDRAWN", "number of items selected", ), )
        if tostd:
            return ( params[0], params[1], params[2], )
        if len(params) != 3:
           raise ValueError("Three parameters expected for the Hypergeometric distribution")
        numtotal = float(params[0])
        numgood = float(params[1])
        numdrawn = float(params[2])
        if (numtotal <= 0.0) or (numgood < 0.0) or (numdrawn < 0.0):
           raise ValueError("Invalid parameter(s) for the Hypergeometric distribution")
        return ( numtotal, numgood, numdrawn, )
    if distname == "invgamma":
        if params is None:
            return ( ( "ALPHA", "shape", ),
                     ( "BETA", "scale", ), )
        if tostd:
            return ( params[0], params[2], params[1], )
        if (len(params) < 2) or (len(params) > 3):
            raise ValueError("Two or three parameters expected for the Inverse-Gamma distribution")
        alpha = float(params[0])
        beta = float(params[1])
        if (alpha <= 0.0) or (beta <= 0.0):
            raise ValueError("Invalid parameter(s) for the Inverse-Gamma distribution")
        try:
            offset = float(params[2])
        except IndexError:
            offset = 0.0
        return ( alpha, offset, beta, )
    if distname == "laplace":
        if params is None:
            return ( ( "MU", "location (mean)", ),
                     ( "B", "scale", ), )
        if tostd:
            return ( params[0], params[1], )
        if len(params) != 2:
            raise ValueError("Two parameters expected for the Laplace distribution")
        mu = float(params[0])
        b = float(params[1])
        if b <= 0.0:
            raise ValueError("Invalid parameters for the Laplace distribution")
        return ( mu, b, )
    if distname == "lognorm":
        if params is None:
            return ( ( "MU", "log-scale (mean of the natural log of the distribution)", ),
                     ( "SIGMA", "shape (std. dev. of the natural log of the distribution)", ), )
        if tostd:
            return ( math.log(params[2]), params[0], params[1], )
        if (len(params) < 2) or (len(params) > 3):
            raise ValueError("Two or three parameters expected for the Log-Normal distribution")
        mu = math.exp(float(params[0]))
        sigma = float(params[1])
        if sigma <= 0.0:
            raise ValueError("Invalid parameter for the Log-Normal distribution")
        try:
            offset = float(params[2])
        except IndexError:
            offset = 0.0
        return ( sigma, offset, mu, )
    if distname == "nbinom":
        if params is None:
            return ( ( "N", "number of successes to stop", ),
                     ( "P", "success probability in each trial", ), )
        if tostd:
            return ( params[0], params[1], )
        if len(params) != 2:
            raise ValueError("Two parameters expected for the Negative-Binomial distribution")
        numsuccess = float(params[0])
        prob = float(params[1])
        if (numsuccess < 1.0) or (prob <= 0.0) or (prob > 1.0):
            raise ValueError("Invalid parameter(s) for the Negative-Binomial distribution")
        return ( numsuccess, prob, )
    if distname == "norm":
        if params is None:
            return ( ( "MU", "mean value", ),
                     ( "SIGMA", "standard deviation", ), )
        if tostd:
            return ( params[0], params[1], )
        if len(params) != 2:
            raise ValueError("Two parameters expected for the Normal distribution")
        mu = float(params[0])
        sigma = float(params[1])
        if sigma <= 0.0:
            raise ValueError("Invalid parameter for the Normal distribution")
        return ( mu, sigma, )
    if distname == "pareto":
        if params is None:
            return ( ( "XM", "scale (minimum abscissa value)", ),
                     ( "ALPHA", "shape", ), )
        if tostd:
            return ( params[2], params[0], params[1], )
        if (len(params) < 2) or (len(params) > 3):
            raise ValueError("Two or three parameters expected for the Pareto distribution")
        xm =  float(params[0])
        alpha = float(params[1])
        if (xm <= 0.0) or (alpha <= 0.0):
            raise ValueError("Invalid parameter(s) for the Pareto distribution")
        try:
            offset = float(params[2])
        except IndexError:
            offset = 0.0
        return ( alpha, offset, xm, )
    if distname == "poisson":
        if params is None:
            return ( ( "MU", "expected number of occurences", ), )
        if tostd:
            return ( params[0], )
        if len(params) != 1:
            raise ValueError("One parameter expected for the Poisson distribution")
        mu = float(params[0])
        if mu <= 0.0:
            raise ValueError("Invalid parameter for the Poisson distribution")
        return ( mu, )
    if distname == "randint":
        if params is None:
            return ( ( "MIN", "minimum integer", ),
                     ( "MAX", "maximum integer (included)", ), )
        if tostd:
            return ( params[0], params[1] - 1, )
        if len(params) != 2:
            raise ValueError("Two parameters expected for the Random-Integer distribution")
        min = int(params[0])
        max = int(params[1])
        # randint takes int values, thus float values are truncated
        # this could lead to unexpected behavior (eg, one might expect
        # (0.9,10.1) to be treated as [1,11) but instead it becomes [0,10)
        minflt = float(params[0])
        maxflt = float(params[1])
        if (min >= max) or (min != minflt) or (max != maxflt):
            raise ValueError("Invalid parameters for the Random-Integer distribution")
        return ( min, max + 1, )
    if distname == "t":
        if params is None:
            return ( ( "DF", "degrees of freedom", ), )
        if tostd:
            return ( params[0], params[1], params[2], )
        if (len(params) < 1) or (len(params) > 3):
            raise ValueError("One to three parameters expected for the Students-T distribution")
        degfree = float(params[0])
        if degfree <= 0.0:
            raise ValueError("Invalid parameter for the Students-T distribution")
        offset = 0.0
        scaling = 1.0
        try:
            offset = float(params[1])
            scaling = float(params[2])
        except IndexError:
            pass
        return ( degfree, offset, scaling, )
    if distname == "uniform":
        if params is None:
            return ( ( "MIN", "minimum", ),
                     ( "MAX", "maximum", ), )
        if tostd:
            return ( params[0], params[0] + params[1], )
        if len(params) != 2:
            raise ValueError("Two parameters expected for the Uniform distribution")
        min = float(params[0])
        max = float(params[1])
        if min >= max:
            raise ValueError("Invalid parameters for the Uniform distribution")
        # these are the "loc" and "scale" parameters for the uniform distribution
        return ( min, max - min, )
    if distname == "weibull_min":
        if params is None:
            return ( ( "K", "shape", ),
                     ( "LAMBDA", "scale", ), )
        if tostd:
            return ( params[0], params[2], params[1], )
        if (len(params) < 2) or (len(params) > 3):
            raise ValueError("Two or three parameters expected for the Weibull distribution")
        k =  float(params[0])
        lambdaflt = float(params[1])
        if (k <= 0.0) or (lambdaflt <= 0.0):
            raise ValueError("Invalid parameter(s) for the Weibull distribution")
        try:
            offset = float(params[2])
        except IndexError:
            offset = 0.0
        return ( k, offset, lambdaflt, )
    return None


def getdistrib(distribname, distribparams):
    """
    Creates and returns scipy.stats "frozen" probability distribution
    object.  Converts the "standard" parameters (including ordering)
    for a distribution to appropriate parameters for calling the
    constructor of the scipy.stats frozen distribution object.

    Arguments:
       distribname - name of the probability distribution
       distribparams - tuple/list/array of standard input parameters

    Returns:
       the scipy.stats "frozen" probability distribution object
           described by distribname and distribparams

    Raises:
       ValueError if the distribution name is not recognized by this
                  routine or if the distribution parameters are invalid
    """
    if (distribname is None) or (distribparams is None):
        raise ValueError("Neither distribname nor distribparams can be None")
    distscipyname = getdistname(distribname)
    if distscipyname is None:
        raise ValueError("Unknown probability function %s" % str(distribname))
    distscipyparams = getdistparams(distscipyname, distribparams)
    if distscipyparams is None:
        raise ValueError("Unknown (for params) probability function %s" % str(distribname))
    distfunc = eval("scipy.stats.%s" % distscipyname)
    distrib = distfunc(*distscipyparams)
    return distrib


def getfitparams(values, distribname, estparams):
    """
    Returns a tuple of "standard" parameters (including ordering) for a
    continuous probability distribution type named in distribname that
    best fits the distribution of data given in values (a 1-D array of
    data with no missing values).  Initial estimates for these "standard"
    parameters are given in estparams.
    """
    if (distribname is None) or (estparams is None):
        raise ValueError("Neither distribname nor estparams can be None")
    distscipyname = getdistname(distribname)
    if distscipyname is None:
        raise ValueError("Unknown probability function %s" % str(distribname))
    estscipyparams = getdistparams(distscipyname, estparams)
    if estscipyparams is None:
        raise ValueError("Unknown (for params) probability function %s" % str(distribname))
    try:
        fitfunc = eval("scipy.stats.%s.fit" % distscipyname)
    except AttributeError:
        raise ValueError("No fit function for probability function %s" % str(distribname))
    if distscipyname == "uniform":
        # "params" keyword for the uniform distribution does not work as expected
        fitscipyparams = fitfunc(values, loc=estscipyparams[0], scale=estscipyparams[1])
    elif distscipyname == "gamma":
        # "params" keyword for the gamma distribution does not work as expected
        fitscipyparams = fitfunc(values, loc=estscipyparams[1], scale=estscipyparams[0])
    elif distscipyname == "laplace":
        # "params" keyword for the laplace distribution does not work as expected
        fitscipyparams = fitfunc(values, loc=estscipyparams[0], scale=estscipyparams[1])
    elif distscipyname == "t":
        # "params" keyword for the student-t test distribution does not work as expected
        fitscipyparams = fitfunc(values, loc=estscipyparams[0], scale=estscipyparams[1])
    else:
        fitscipyparams = fitfunc(values, params=estscipyparams)
    return getdistparams(distscipyname, fitscipyparams, tostd=True)


def getinitdict(distribname, funcname):
    """
    Returns a dictionary appropriate for the return value of ferret_init
    in a Ferret stats_<disribname>_<funcname> PyEF

    Arguments:
       distribname - name of the probability distribution
       funcname - name of the scipy.stats probability distribution function
    """
    # generate a long function name from the scipy.stats function name
    if ( funcname == "cdf" ):
        funcaction = "calculate"
        funcreturn = "cumulative density function values"
    elif ( funcname == "isf" ):
        funcaction = "calculate"
        funcreturn = "inversion survival function values"
    elif ( funcname == "pdf" ):
        funcaction = "calculate"
        funcreturn = "probability distribution function values"
    elif ( funcname == "pmf" ):
        funcaction = "calculate"
        funcreturn = "probability mass function values"
    elif ( funcname == "ppf" ):
        funcaction = "calculate"
        funcreturn = "percent point function values"
    elif ( funcname == "sf" ):
        funcaction = "calculate"
        funcreturn = "survival function values"
    elif ( funcname == "rvs" ):
        funcaction = "assign"
        funcreturn = "random variates"
    else:
        raise ValueError("Unsupported scipy.stats function name '%s'" % funcname)
    # Get the distribution parameters information
    distscipyname = getdistname(distribname)
    paramdescripts = getdistparams(distscipyname, None)
    numargs = len(paramdescripts) + 1
    descript = "Returns array of %s for %s prob. distrib." % (funcreturn, distribname)
    axes = [ pyferret.AXIS_IMPLIED_BY_ARGS ] * pyferret.MAX_FERRET_NDIM
    argtypes = [ pyferret.FLOAT_ARRAY ] * numargs
    influences = [ [ True ] * pyferret.MAX_FERRET_NDIM ] * numargs
    if ( numargs == 2 ):
        # info for distributions with one parameter
        argnames = ( "PTS", paramdescripts[0][0], )
        argdescripts = ( "Point(s) at which to %s the %s" % (funcaction, funcreturn),
                         "Parameter(s) defining the %s" % paramdescripts[0][1], )
    elif (numargs == 3):
        # info for distributions with two parameters
        argnames = ( "PTS", paramdescripts[0][0], paramdescripts[1][0], )
        argdescripts = ( "Point(s) at which to %s the %s" % (funcaction, funcreturn),
                         "Parameter(s) defining the %s" % paramdescripts[0][1],
                         "Parameter(s) defining the %s" % paramdescripts[1][1], )
    elif (numargs == 4):
        # info for distributions with three parameters
        argnames = ( "PTS", paramdescripts[0][0], paramdescripts[1][0], paramdescripts[2][0], )
        argdescripts = ( "Point(s) at which to %s the %s" % (funcaction, funcreturn),
                         "Parameter(s) defining the %s" % paramdescripts[0][1],
                         "Parameter(s) defining the %s" % paramdescripts[1][1],
                         "Parameter(s) defining the %s" % paramdescripts[2][1], )
    else:
        raise ValueError("Unexpected number of arguments: %d" % numargs)
    # Create and return the dictionary
    return { "numargs": numargs,
             "descript": descript,
             "axes": axes,
             "argnames": argnames,
             "argdescripts": argdescripts,
             "argtypes": argtypes,
             "influences": influences, }


def getdistribfunc(distrib, funcname):
    """
    Returns the distrib.funcname function for recognized funcnames
    """
    if ( funcname == "cdf" ):
        return distrib.cdf
    elif ( funcname == "isf" ):
        return distrib.isf
    elif ( funcname == "pdf" ):
        return distrib.pdf
    elif ( funcname == "pmf" ):
        return distrib.pmf
    elif ( funcname == "ppf" ):
        return distrib.ppf
    elif ( funcname == "sf" ):
        return distrib.sf
    elif ( funcname == "rvs" ):
        return distrib.rvs
    else:
        raise ValueError("Unsupported scipy.stats function name '%s'" % funcname)


def assignresultsarray(distribname, funcname, result, resbdf, inputs, inpbdfs):
    """
    Assigns result with the funcname function values for the distribname
    probability distributions defined by parameters in inputs[1:]
    using the abscissa or template values given in inputs[0].
    """
    # get the masks for the PTS data
    pts = inputs[0]
    badmask = ( numpy.fabs(pts - inpbdfs[0]) < 1.0E-7 )
    badmask = numpy.logical_or(badmask, numpy.isnan(pts))
    goodmask = numpy.logical_not(badmask)
    # figure out the axes for the parameters
    numparams = len(inputs) - 1
    if numparams > 0:
        par1axis = -1
        for k in range(pyferret.MAX_FERRET_NDIM):
            if inputs[1].shape[k] > 1:
                if par1axis != -1:
                    raise ValueError("Parameters arrays can have only one defined, non-singular axis")
                par1axis = k
        if par1axis >= 0:
            if pts.shape[par1axis] > 1:
                raise ValueError("Unexpected error: shape[%d] of PTS and PAR1 both > 1" % par1axis)
        par1vals = inputs[1].reshape(-1)
        # temporary results array for a given first parameter
        tmp1result = numpy.empty(pts.shape, dtype=numpy.float64, order='F')
        tmp1result[badmask] = resbdf
    if numparams > 1:
        par2axis = -1
        for k in range(pyferret.MAX_FERRET_NDIM):
            if inputs[2].shape[k] > 1:
                if par2axis != -1:
                    raise ValueError("Parameters arrays can have only one defined, non-singular axis")
                par2axis = k
        if par2axis >= 0:
            if pts.shape[par2axis] > 1:
                raise ValueError("Unexpected error: shape[%d] of PTS and PAR2 both > 1" % par2axis)
            if par1axis == par2axis:
                raise ValueError("Unexpected error: shape[%d] of PAR1 and PAR2 both > 1" % par2axis)
        par2vals = inputs[2].reshape(-1)
        # temporary results array for a given second parameter and all of the first parameters
        if par1axis == -1:
            tmp2result = tmp1result
        else:
            shape = list(tmp1result.shape)
            shape[par1axis] = len(par1vals)
            tmp2result = numpy.empty(shape, dtype=numpy.float64, order='F')
    if numparams > 2:
        par3axis = -1
        for k in range(pyferret.MAX_FERRET_NDIM):
            if inputs[3].shape[k] > 1:
                if par3axis != -1:
                    raise ValueError("Parameters arrays can have only one defined, non-singular axis")
                par3axis = k
        if par3axis >= 0:
            if pts.shape[par3axis] > 1:
                raise ValueError("Unexpected error: shape[%d] of PTS and PAR3 both > 1" % par3axis)
            if par1axis == par3axis:
                raise ValueError("Unexpected error: shape[%d] of PAR1 and PAR3 both > 1" % par3axis)
            if par2axis == par3axis:
                raise ValueError("Unexpected error: shape[%d] of PAR2 and PAR3 both > 1" % par3axis)
        par3vals = inputs[3].reshape(-1)
        # temporary results array for a given third parameter and all of the first and second parameters
        if par2axis == -1:
            tmp3result = tmp2result
        else:
            shape = list(tmp2result.shape)
            shape[par2axis] = len(par2vals)
            tmp3result = numpy.empty(shape, dtype=numpy.float64, order='F')
    # deal with each number of parameters separately when assigning results
    if numparams == 1:
        for j in range(len(par1vals)):
            try:
                if math.isnan(par1vals[j]) or (math.fabs(par1vals[j] - inpbdfs[1]) < 1.0E-7):
                    raise ValueError
                distrib = getdistrib(distribname, ( par1vals[j], ))
                distribfunc = getdistribfunc(distrib, funcname)
                if funcname == "rvs":
                    tmp1result[goodmask] = getdistribfunc(distrib, funcname)(len(pts[goodmask]))
                else:
                    tmp1result[goodmask] = getdistribfunc(distrib, funcname)(pts[goodmask])
            except ValueError:
                tmp1result[goodmask] = resbdf
            # Appropriately assign the tmp1result to result
            if par1axis == -1:
                result[:, :, :, :, :, :] = tmp1result
            elif par1axis == 0:
                result[j, :, :, :, :, :] = tmp1result[0, :, :, :, :, :]
            elif par1axis == 1:
                result[:, j, :, :, :, :] = tmp1result[:, 0, :, :, :, :]
            elif par1axis == 2:
                result[:, :, j, :, :, :] = tmp1result[:, :, 0, :, :, :]
            elif par1axis == 3:
                result[:, :, :, j, :, :] = tmp1result[:, :, :, 0, :, :]
            elif par1axis == 4:
                result[:, :, :, :, j, :] = tmp1result[:, :, :, :, 0, :]
            elif par1axis == 5:
                result[:, :, :, :, :, j] = tmp1result[:, :, :, :, :, 0]
            else:
                raise ValueError("Unexpected par1axis of %d" % par1axis)
    elif numparams == 2:
        for k in range(len(par2vals)):
            for j in range(len(par1vals)):
                try:
                    if math.isnan(par2vals[k]) or (math.fabs(par2vals[k] - inpbdfs[2]) < 1.0E-7) or \
                       math.isnan(par1vals[j]) or (math.fabs(par1vals[j] - inpbdfs[1]) < 1.0E-7):
                        raise ValueError
                    distrib = getdistrib(distribname, ( par1vals[j], par2vals[k], ))
                    distribfunc = getdistribfunc(distrib, funcname)
                    if funcname == "rvs":
                        tmp1result[goodmask] = getdistribfunc(distrib, funcname)(len(pts[goodmask]))
                    else:
                        tmp1result[goodmask] = getdistribfunc(distrib, funcname)(pts[goodmask])
                except ValueError:
                    tmp1result[goodmask] = resbdf
                # Appropriately assign the tmp1result to tmp2result
                if par1axis == -1:
                    # in this case tmp2result is tmp1result
                    pass
                elif par1axis == 0:
                    tmp2result[j, :, :, :, :, :] = tmp1result[0, :, :, :, :, :]
                elif par1axis == 1:
                    tmp2result[:, j, :, :, :, :] = tmp1result[:, 0, :, :, :, :]
                elif par1axis == 2:
                    tmp2result[:, :, j, :, :, :] = tmp1result[:, :, 0, :, :, :]
                elif par1axis == 3:
                    tmp2result[:, :, :, j, :, :] = tmp1result[:, :, :, 0, :, :]
                elif par1axis == 4:
                    tmp2result[:, :, :, :, j, :] = tmp1result[:, :, :, :, 0, :]
                elif par1axis == 5:
                    tmp2result[:, :, :, :, :, j] = tmp1result[:, :, :, :, :, 0]
                else:
                    raise ValueError("Unexpected par1axis of %d" % par1axis)
            # Appropriately assign the tmp2result to result
            if par2axis == -1:
                result[:, :, :, :, :, :] = tmp2result
            elif par2axis == 0:
                result[k, :, :, :, :, :] = tmp2result[0, :, :, :, :, :]
            elif par2axis == 1:
                result[:, k, :, :, :, :] = tmp2result[:, 0, :, :, :, :]
            elif par2axis == 2:
                result[:, :, k, :, :, :] = tmp2result[:, :, 0, :, :, :]
            elif par2axis == 3:
                result[:, :, :, k, :, :] = tmp2result[:, :, :, 0, :, :]
            elif par2axis == 4:
                result[:, :, :, :, k, :] = tmp2result[:, :, :, :, 0, :]
            elif par2axis == 5:
                result[:, :, :, :, :, k] = tmp2result[:, :, :, :, :, 0]
            else:
                raise ValueError("Unexpected par2axis of %d" % par2axis)
    elif numparams == 3:
        for q in range(len(par3vals)):
            for k in range(len(par2vals)):
                for j in range(len(par1vals)):
                    try:
                        if math.isnan(par3vals[q]) or (math.fabs(par3vals[q] - inpbdfs[3]) < 1.0E-7) or \
                           math.isnan(par2vals[k]) or (math.fabs(par2vals[k] - inpbdfs[2]) < 1.0E-7) or \
                           math.isnan(par1vals[j]) or (math.fabs(par1vals[j] - inpbdfs[1]) < 1.0E-7):
                            raise ValueError
                        distrib = getdistrib(distribname, ( par1vals[j], par2vals[k], par3vals[q], ))
                        distribfunc = getdistribfunc(distrib, funcname)
                        if funcname == "rvs":
                            tmp1result[goodmask] = getdistribfunc(distrib, funcname)(len(pts[goodmask]))
                        else:
                            tmp1result[goodmask] = getdistribfunc(distrib, funcname)(pts[goodmask])
                    except ValueError:
                        tmp1result[goodmask] = resbdf
                    # Appropriately assign the tmp1result to tmp2result
                    if par1axis == -1:
                        # in this case tmp2result is tmp1result
                        pass
                    elif par1axis == 0:
                        tmp2result[j, :, :, :, :, :] = tmp1result[0, :, :, :, :, :]
                    elif par1axis == 1:
                        tmp2result[:, j, :, :, :, :] = tmp1result[:, 0, :, :, :, :]
                    elif par1axis == 2:
                        tmp2result[:, :, j, :, :, :] = tmp1result[:, :, 0, :, :, :]
                    elif par1axis == 3:
                        tmp2result[:, :, :, j, :, :] = tmp1result[:, :, :, 0, :, :]
                    elif par1axis == 4:
                        tmp2result[:, :, :, :, j, :] = tmp1result[:, :, :, :, 0, :]
                    elif par1axis == 5:
                        tmp2result[:, :, :, :, :, j] = tmp1result[:, :, :, :, :, 0]
                    else:
                        raise ValueError("Unexpected par1axis of %d" % par1axis)
                # Appropriately assign the tmp2result to tmp3result
                if par2axis == -1:
                    # in this case tmp3result is tmp2result
                    pass
                elif par2axis == 0:
                    tmp3result[k, :, :, :, :, :] = tmp2result[0, :, :, :, :, :]
                elif par2axis == 1:
                    tmp3result[:, k, :, :, :, :] = tmp2result[:, 0, :, :, :, :]
                elif par2axis == 2:
                    tmp3result[:, :, k, :, :, :] = tmp2result[:, :, 0, :, :, :]
                elif par2axis == 3:
                    tmp3result[:, :, :, k, :, :] = tmp2result[:, :, :, 0, :, :]
                elif par2axis == 4:
                    tmp3result[:, :, :, :, k, :] = tmp2result[:, :, :, :, 0, :]
                elif par2axis == 5:
                    tmp3result[:, :, :, :, :, k] = tmp2result[:, :, :, :, :, 0]
                else:
                    raise ValueError("Unexpected par2axis of %d" % par2axis)
            # Appropriately assign the tmp3result to result
            if par3axis == -1:
                result[:, :, :, :, :, :] = tmp3result
            elif par3axis == 0:
                result[q, :, :, :, :, :] = tmp3result[0, :, :, :, :, :]
            elif par3axis == 1:
                result[:, q, :, :, :, :] = tmp3result[:, 0, :, :, :, :]
            elif par3axis == 2:
                result[:, :, q, :, :, :] = tmp3result[:, :, 0, :, :, :]
            elif par3axis == 3:
                result[:, :, :, q, :, :] = tmp3result[:, :, :, 0, :, :]
            elif par3axis == 4:
                result[:, :, :, :, q, :] = tmp3result[:, :, :, :, 0, :]
            elif par3axis == 5:
                result[:, :, :, :, :, q] = tmp3result[:, :, :, :, :, 0]
            else:
                raise ValueError("Unexpected par3axis of %d" % par3axis)
    else:
        raise ValueError("Unexpected number of parameters: %d" % numparams)


#
# The rest of this is just for testing this module at the command line
#
if __name__ == "__main__":
    # Test getdistname names
    namelist = getdistname(None)
    if len(namelist) < 22:
        raise ValueError("Too few of distributions: expected at least 22; found %d" % \
                         len(distdescripts))
    for nametuple in namelist:
        for name in nametuple:
            statsname = getdistname(name)
            if statsname != nametuple[0]:
                raise ValueError("getdistname(%s): expected %s; found %s" % \
                                 (name, nametuple[0], statsname))

    # Test the distribution scipy name and parameters given to getdistrib
    # give the expected distribution.  (Primarily that the parameters
    # are interpreted and assigned correctly.)
    # Testing of the long names is performed by the stats_helper.py script.
    # Testing of assignresultsarray is done in the stats_norm_*.py scipts.

    # Beta distribution
    distname = "beta"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d" % \
                         (distname, len(descript)))
    del descript

    alpha = 1.5
    beta = 2.75
    distparms = [ alpha, beta ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( alpha / (alpha + beta),
                      alpha * beta / ((alpha + beta)**2 * (alpha + beta + 1.0)),
                      2.0 * (beta - alpha) / (2.0 + alpha + beta) *
                          math.sqrt((1.0 + alpha + beta) / (alpha * beta)),
                      6.0 * (alpha**3 + alpha**2 * (1.0 - 2.0 * beta) + \
                          beta**2 * (1.0 + beta) - 2.0 * alpha * beta * (2.0 + beta)) / \
                          (alpha * beta * (alpha + beta + 2.0) * (alpha + beta + 3.0)),
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del alpha, beta, foundstats, expectedstats

    # append the default loc and scale to the expected params
    distparms.append(0.0)
    distparms.append(1.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # Binomial distribution
    distname = "binom"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    ntrials = 20.0
    prob = 0.25
    distparms = [ ntrials, prob ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( ntrials * prob,
                      ntrials * prob * (1.0 - prob),
                      (1.0 - 2.0 * prob) / math.sqrt(ntrials * prob * (1.0 - prob)),
                      (1.0 - 6.0 * prob * (1.0 - prob)) / (ntrials * prob * (1.0 - prob)),
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del ntrials, prob, foundstats, expectedstats

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # no binom.fit function
    del distparms, distf, newparms

    print("%s: PASS" % distname)


    # Cauchy distribution
    distname = "cauchy"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    m = 5.0
    gamma = 2.0
    distparms = [ m, gamma ]
    distf = getdistrib(distname, distparms)
    # mean, variance, skew, kurtosis undefined; instead check some pdf values
    xvals = numpy.arange(0.0, 10.1, 0.5)
    foundpdfs = distf.pdf(xvals)
    expectedpdfs = (gamma / numpy.pi) / ((xvals - m)**2 + gamma**2)
    if not numpy.allclose(foundpdfs, expectedpdfs):
        print("%s: FAIL" % distname)
        raise ValueError("pdfs(0.0:10.1:0.5) of %s(%#.1f,%#.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedpdfs), str(foundpdfs)))
    del m, gamma, xvals, foundpdfs, expectedpdfs

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # Chi distribution
    distname = "chi"
    descript = getdistparams(distname, None)
    if len(descript) != 1:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 1; found %d:" % \
                         (distname, len(descript)))
    del descript

    degfree = 10
    distparms = [ degfree ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    mean = math.sqrt(2.0) * scipy.special.gamma(0.5 * (degfree + 1.0)) / \
                            scipy.special.gamma(0.5 * degfree)
    variance = degfree - mean**2
    stdev = math.sqrt(variance)
    skew = mean * (1.0 - 2.0 * variance) / stdev**3
    expectedstats = ( mean,
                      variance,
                      skew,
                      2.0 * (1.0 - mean * stdev * skew - variance) / variance,
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%d.0): expected %s; found %s" % \
                          (distname, distparms[0], str(expectedstats), str(foundstats)))
    del degfree, foundstats, mean, variance, stdev, skew, expectedstats

    # append the default loc and scale to the expected params
    distparms.append(0.0)
    distparms.append(1.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.4, atol=0.4):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # Chi-squared distribution
    distname = "chi2"
    descript = getdistparams(distname, None)
    if len(descript) != 1:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 1; found %d:" % \
                         (distname, len(descript)))
    del descript

    degfreestr = "10"
    distparms = [ degfreestr ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    degfree = float(degfreestr)
    expectedstats = ( degfree,
                      2.0 * degfree,
                      math.sqrt(8.0 / degfree),
                      12.0 / degfree,
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%s): expected %s; found %s" % \
                          (distname, distparms[0], str(expectedstats), str(foundstats)))
    del degfreestr, foundstats, expectedstats

    # append the default loc and scale to the expected (numeric) params
    distparms = [ degfree, 0.0, 1.0 ]
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.4, atol=0.4):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del degfree, distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # Exponential distribution
    distname = "expon"
    descript = getdistparams(distname, None)
    if len(descript) != 1:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 1; found %d:" % \
                         (distname, len(descript)))
    del descript

    lambdaflt = 11.0
    distparms = [ lambdaflt ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( 1.0 / lambdaflt, 1.0 / lambdaflt**2, 2.0, 6.0 )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f): expected %s; found %s" % \
                          (distname, distparms[0], str(expectedstats), str(foundstats)))
    del lambdaflt, foundstats, expectedstats

    # append the default loc to the expected params
    distparms.append(0.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # Exponentiated Weibull distribution
    distname = "exponweib"
    descript = getdistparams(distname, None)
    if len(descript) != 3:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 3; found %d:" % \
                         (distname, len(descript)))
    del descript

    k = 3.0
    lambdaflt = 5.0
    alpha = 2.5
    distparms = [ k, lambdaflt, alpha ]
    distf = getdistrib(distname, distparms)
    # don't know the formula for the mean, variance, skew, kurtosis
    # instead check some cdf values
    xvals = numpy.arange(0.0, 10.1, 0.5)
    foundcdfs = distf.cdf(xvals)
    expectedcdfs = numpy.power(1.0 - numpy.exp(-1.0 * numpy.power(xvals / lambdaflt, k)), alpha)
    if not numpy.allclose(foundcdfs, expectedcdfs):
        print("%s: FAIL" % distname)
        raise ValueError("cdfs(0.0:10.1:0.5) of %s(%#.1f,%#.1f%#.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], distparms[2], str(expectedcdfs), str(foundcdfs)))
    del k, lambdaflt, alpha, xvals, foundcdfs, expectedcdfs

    # append the default loc to the expected params
    distparms.append(0.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # F distribution
    distname = "f"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    dofn = 7.0
    dofd = 11.0   # needs to be larger than 8.0 for kurtosis formula
    distparms = [ dofn, dofd ]
    distf = getdistrib(distname, distparms)
    # foundstats = distf.stats("mvsk")
    foundstats = distf.stats("mv")
    expectedstats = ( dofd / (dofd - 2.0),
                      2.0 * dofd**2 * (dofn + dofd - 2.0) / \
                          (dofn * (dofd - 2.0)**2 * (dofd - 4.0)),
                      # ((2.0 * dofn + dofd - 2.0) / (dofd - 6.0)) * \
                      #     math.sqrt(8.0 * (dofd - 4.0) / (dofn * (dofn + dofd - 2.0))),
                      # 12.0 * (20.0 * dofd - 8.0 * dofd**2 + dofd**3 + 44.0 * dofn - 32.0 * dofn * dofd + \
                      #     5.0 * dofd**2 * dofn - 22.0 * dofn**2 + 5.0 * dofd * dofn**2 - 16.0) / \
                      #     (dofn * (dofd - 6.0) * (dofd - 8.0) * (dofn + dofd - 2)),
                    )
    if not numpy.allclose(foundstats, expectedstats):
        # raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    # since skew and kurtosis is not coming out as expected, check some pdf values
    xvals = numpy.arange(0.5, 10.1, 0.5)
    foundpdfs = distf.pdf(xvals)
    factor = scipy.special.gamma(0.5 * (dofn + dofd)) / \
             (scipy.special.gamma(0.5 * dofn) * scipy.special.gamma(0.5 *dofd))
    factor *= math.pow(dofn, 0.5 * dofn) * math.pow(dofd, 0.5 * dofd)
    expectedpdfs = factor * numpy.power(xvals, 0.5 * dofn - 1.0) / \
                   numpy.power(dofd + dofn * xvals, 0.5 * (dofn + dofd))
    if not numpy.allclose(foundpdfs, expectedpdfs):
        print("%s: FAIL" % distname)
        raise ValueError("pdfs(0.5:10.1:0.5) of %s(%#.1f,%#.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedpdfs), str(foundpdfs)))
    del dofn, dofd, foundstats, expectedstats, xvals, foundpdfs, factor, expectedpdfs

    # append the default loc and scale to the expected params
    distparms.append(0.0)
    distparms.append(1.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.2, atol=0.4):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # Gamma distribution
    distname = "gamma"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    alpha = 5.0
    theta = 3.0
    distparms = [ alpha, theta ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( alpha * theta, alpha * theta**2, 2.0 / math.sqrt(alpha), 6.0 / alpha )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del alpha, theta, foundstats, expectedstats

    # append the default loc to the expected params
    distparms.append(0.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    sample = distf.rvs(25000)
    fitparms = getfitparams(sample, distname, distparms)
    if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
        print("%s: FAIL" % distname)
        raise ValueError("fitparams of %s: expected %s; found %s" % \
                                      (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms, sample, fitparms

    print("%s: PASS" % distname)


    # Geometric distribution
    distname = "geom"
    descript = getdistparams(distname, None)
    if len(descript) != 1:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 1; found %d:" % \
                         (distname, len(descript)))
    del descript

    prob = 0.25
    distparms = [ prob ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( 1.0 / prob,
                     (1.0 - prob) / prob**2,
                     (2.0 - prob) / math.sqrt(1.0 - prob),
                     6.0 + prob**2 / (1.0 - prob),
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f): expected %s; found %s" % \
                          (distname, distparms[0], str(expectedstats), str(foundstats)))
    del prob, foundstats, expectedstats

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # no geom.fit function
    del distparms, distf, newparms

    print("%s: PASS" % distname)


    # Hypergeometric distribution
    distname = "hypergeom"
    descript = getdistparams(distname, None)
    if len(descript) != 3:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 3; found %d:" % \
                         (distname, len(descript)))
    del descript

    numtotal = 29.0
    numgood = 13.0
    numdrawn = 17.0
    distparms = [ numtotal, numgood, numdrawn ]
    distf = getdistrib(distname, distparms)
    # foundstats = distf.stats("mvsk")
    foundstats = distf.stats("mvs")
    expectedstats = ( numdrawn * numgood / numtotal,
                      numdrawn * numgood * (numtotal - numdrawn) * (numtotal - numgood) / \
                          (numtotal**2 * (numtotal - 1.0)),
                      math.sqrt(numtotal - 1.0) * (numtotal - 2.0 * numdrawn) * (numtotal - 2.0 * numgood) / \
                          (math.sqrt(numdrawn * numgood * (numtotal - numdrawn) * (numtotal - numgood)) * \
                              (numtotal - 2.0)),
                      # (numtotal**2 * (numtotal - 1.0) / \
                      #         (numdrawn * (numtotal - 2.0) * (numtotal - 3.0) * (numtotal - numdrawn))) * \
                      #     ((numtotal * (numtotal + 1.0) - 6.0 * numtotal * (numtotal - numdrawn)) / \
                      #      (numgood * (numtotal - numgood)) + \
                      #      3.0 * numdrawn * (numtotal - numdrawn) * (numtotal + 6.0) / numtotal**2 - 6.0),
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], distparms[2], str(expectedstats), str(foundstats)))
    del numtotal, numgood, numdrawn, foundstats, expectedstats

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # no hypergeom.fit function
    del distparms, distf, newparms

    print("%s: PASS" % distname)


    # Inverse-Gamma distribution
    distname = "invgamma"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    alpha = 7.0  # must be > 4 for the kurtosis formula
    beta = 3.0
    distparms = [ alpha, beta ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( beta / (alpha - 1.0),
                      beta**2 / ((alpha - 1.0)**2 * (alpha - 2.0)),
                      4.0 * math.sqrt(alpha - 2.0) / (alpha - 3.0),
                      (30.0 * alpha - 66.0)/ ((alpha - 3.0) * (alpha - 4.0)),
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del alpha, beta, foundstats, expectedstats

    # append the default loc to the expected params
    distparms.append(0.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.2, atol=0.4):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # Laplace distribution
    distname = "laplace"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    mu = 5.0
    b = 3.0
    distparms = [ mu, b ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( mu, 2.0 * b**2, 0.0, 3.0 )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del mu, b, foundstats, expectedstats

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    sample = distf.rvs(25000)
    fitparms = getfitparams(sample, distname, distparms)
    if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
        print("%s: FAIL" % distname)
        raise ValueError("fitparams of %s: expected %s; found %s" % \
                                      (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms, sample, fitparms

    print("%s: PASS" % distname)


    # Log-normal distribution
    distname = "lognorm"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    mu = 0.8
    sigma = 0.5
    distparms = [ mu, sigma ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( math.exp(mu + 0.5 * sigma**2),
                      math.exp(2.0 * mu + sigma**2) * (math.exp(sigma**2) - 1.0),
                      (2.0 + math.exp(sigma**2)) * math.sqrt(math.exp(sigma**2) - 1.0),
                      math.exp(4.0 * sigma**2) + 2.0 * math.exp(3.0 * sigma**2) + 3.0 * math.exp(2.0 * sigma**2) - 6,
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del mu, sigma, foundstats, expectedstats

    # append the default loc to the expected params
    distparms.append(0.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # Negative-binomial distribution
    distname = "nbinom"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    numsuccess = 5.0
    prob = 0.25
    distparms = [ numsuccess, prob ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( numsuccess * (1.0 - prob) / prob,
                      numsuccess * (1.0 - prob) / prob**2,
                      (2.0 - prob) / math.sqrt(numsuccess * (1.0 - prob)),
                      (prob**2 - 6.0 * prob + 6.0) / (numsuccess * (1.0 - prob)),
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del numsuccess, prob, foundstats, expectedstats

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # no nbinom.fit function
    del distparms, distf, newparms

    print("%s: PASS" % distname)


    # Normal distribution
    distname = "norm"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    mu = 5.0
    sigma = 3.0
    distparms = numpy.array([ mu, sigma ], dtype=numpy.float64)
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( mu, sigma**2, 0.0, 0.0 )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del mu, sigma, foundstats, expectedstats

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    sample = distf.rvs(25000)
    fitparms = getfitparams(sample, distname, distparms)
    if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
        print("%s: FAIL" % distname)
        raise ValueError("fitparams of %s: expected %s; found %s" % \
                                      (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms, sample, fitparms

    print("%s: PASS" % distname)


    # Pareto distribution
    distname = "pareto"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    xm = 3.0
    alpha = 5.0  # must be larger than 4 for kurtosis formula
    distparms = [ xm, alpha ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( alpha * xm / (alpha - 1.0),
                      xm**2 * alpha / ((alpha - 1.0)**2 * (alpha - 2.0)),
                      2.0 * ((alpha + 1.0) / (alpha - 3.0)) * math.sqrt((alpha - 2.0) / alpha),
                      6.0 * (alpha**3 + alpha**2 - 6.0 * alpha - 2.0) / \
                          (alpha * (alpha - 3.0) * (alpha - 4.0)),
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del xm, alpha, foundstats, expectedstats

    # append the default loc to the expected params
    distparms.append(0.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # Poisson distribution
    distname = "poisson"
    descript = getdistparams(distname, None)
    if len(descript) != 1:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 1; found %d:" % \
                         (distname, len(descript)))
    del descript

    mu = 7.0
    distparms = [ mu ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( mu, mu, 1.0 / math.sqrt(mu), 1.0 / mu )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f): expected %s; found %s" % \
                          (distname, distparms[0], str(expectedstats), str(foundstats)))
    del mu, foundstats, expectedstats

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # no poisson.fit function
    del distparms, distf, newparms

    print("%s: PASS" % distname)


    # Random Integer (Discrete Uniform) distribution
    distname = "randint"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    a = -5
    b = 13
    distparms = [ a, b ]
    distf = getdistrib(distname, distparms)
    # foundstats = distf.stats("mvsk")
    foundstats = distf.stats("mvs")
    n = b - a + 1
    # expectedstats = ( 0.5 * (a + b), (n**2 - 1.0) / 12.0, 0.0, -6.0 * (n**2 + 1) / (5.0 * (n**2 - 1)) )
    expectedstats = ( 0.5 * (a + b), (n**2 - 1.0) / 12.0, 0.0, )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        # raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
        raise ValueError("(mean, var, skew) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    xvals = numpy.arange(a - 1.0, b + 1.1, 1.0)
    expectedpmfs = numpy.ones((n+2,), dtype=float) / n
    expectedpmfs[0] = 0.0
    expectedpmfs[n+1] = 0.0
    foundpmfs = distf.pmf(xvals)
    if not numpy.allclose(foundpmfs, expectedpmfs):
        print("%s: FAIL" % distname)
        raise ValueError("pmfs(%.1f:%.1f:1.0) of %s(%.1f, %.1f): expected %s; found %s" % \
              (a - 1.0, b + 1.1, distname, distparms[0], distparms[1], str(expectedpmfs), str(foundpmfs)))
    del a, b, foundstats, n, expectedstats, xvals, expectedpmfs, foundpmfs

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # no randint.fit function
    del distparms, distf, newparms

    print("%s: PASS" % distname)


    # Student's-t distribution
    distname = "t"
    descript = getdistparams(distname, None)
    if len(descript) != 1:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 1; found %d:" % \
                         (distname, len(descript)))
    del descript

    degfree = 11.0
    distparms = [ degfree ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( 0.0, degfree / (degfree - 2.0), 0.0, 6.0 / (degfree - 4.0) )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f): expected %s; found %s" % \
                          (distname, distparms[0], str(expectedstats), str(foundstats)))
    del degfree, foundstats, expectedstats

    # append the default loc and scale to the expected params
    distparms.append(0.0)
    distparms.append(1.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    sample = distf.rvs(25000)
    fitparms = getfitparams(sample, distname, distparms)
    if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
        print("%s: FAIL" % distname)
        raise ValueError("fitparams of %s: expected %s; found %s" % \
                                      (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms, sample, fitparms

    print("%s: PASS" % distname)


    # Uniform distribution
    distname = "uniform"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    a = -5.0
    b = 13.0
    distparms = [ a, b ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    expectedstats = ( 0.5 * (a + b), (b - a)**2 / 12.0, 0.0, -6.0 / 5.0 )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del a, b, foundstats, expectedstats

    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    sample = distf.rvs(25000)
    fitparms = getfitparams(sample, distname, distparms)
    if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
        print("%s: FAIL" % distname)
        raise ValueError("fitparams of %s: expected %s; found %s" % \
                                      (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms, sample, fitparms

    print("%s: PASS" % distname)


    # Weibull distribution
    distname = "weibull_min"
    descript = getdistparams(distname, None)
    if len(descript) != 2:
        print("%s: FAIL" % distname)
        raise ValueError("number of parameter description pairs for %s: expected 2; found %d:" % \
                         (distname, len(descript)))
    del descript

    k = 3.0
    lambdaflt = 5.0
    distparms = [ k, lambdaflt ]
    distf = getdistrib(distname, distparms)
    foundstats = distf.stats("mvsk")
    gam1 = scipy.special.gamma(1.0 + 1.0 / k)
    gam2 = scipy.special.gamma(1.0 + 2.0 / k)
    gam3 = scipy.special.gamma(1.0 + 3.0 / k)
    gam4 = scipy.special.gamma(1.0 + 4.0 / k)
    mu = lambdaflt * gam1
    sigma = lambdaflt * math.sqrt(gam2 - gam1**2)
    expectedstats = ( mu,
                      sigma**2,
                      (lambdaflt**3 * gam3  - 3.0 * mu * sigma**2 - mu**3) / sigma**3,
                      (gam4 - 4.0 * gam1 * gam3 - 3.0 * gam2**2 + 12.0 * gam1**2 * gam2 - 6.0 * gam1**4) / \
                      (gam2 - gam1**2)**2,
                    )
    if not numpy.allclose(foundstats, expectedstats):
        print("%s: FAIL" % distname)
        raise ValueError("(mean, var, skew, kurtosis) of %s(%.1f, %.1f): expected %s; found %s" % \
                          (distname, distparms[0], distparms[1], str(expectedstats), str(foundstats)))
    del k, lambdaflt, foundstats, gam1, gam2, gam3, gam4, mu, sigma, expectedstats

    # append the default loc to the expected params
    distparms.append(0.0)
    newparms = getdistparams(distname, getdistparams(distname, distparms, tostd=False), tostd=True)
    if not numpy.allclose(newparms, distparms):
        print("%s: FAIL" % distname)
        raise ValueError("conversion of full %s params to scipy then back to std changed" % distname)
    # sample = distf.rvs(25000)
    # fitparms = getfitparams(sample, distname, distparms)
    # print "%s fitparams: %s" % (distname, str(fitparms))
    # if not numpy.allclose(fitparms, distparms, rtol=0.1, atol=0.2):
    #     print "%s: FAIL" % distname
    #     raise ValueError("fitparams of %s: expected %s; found %s" % \
    #                                   (distname, str(distparms), str(fitparms)))
    del distparms, distf, newparms # , sample, fitparms

    print("%s: PASS" % distname)


    # All successful
    print("Success")




"""
Creates a shapefile with a given root name using data from a given value
array.  The shapes are quadrilaterals in the X,Y-plane at a given Z (if
this axis is used).  The vertices of the quadrilaterals are the values
of the bounding boxes of the X and Y axes of the given value array.  The
value(s) associated with each shape comes from the value array.
"""

from __future__ import print_function

import pyferret
import pyferret.fershp
import shapefile

def ferret_init(efid):
    """
    Initialization for the shapefile_writeval PyEF
    """
    retdict = { "numargs": 4,
                "descript": "Writes a shapefile of XY quadrilaterals associated with given values",
                "restype": pyferret.FLOAT_ARRAY,
                "axes": ( pyferret.AXIS_ABSTRACT,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST,
                          pyferret.AXIS_DOES_NOT_EXIST, ),
                "argnames": ( "SHAPEFILE", "VALUE", "VALNAME", "MAPPRJ"),
                "argdescripts": ( "Shapefile name (any extension given is ignored)",
                                  "Shape values; X and Y axes required, " \
                                      "Z axis optional, T axis undefined or singleton",
                                  "Name for the shape value",
                                  "Common name or WKT description of map projection; " \
                                      "if blank, WGS 84 is used", ),
                "argtypes": ( pyferret.STRING_ONEVAL,
                              pyferret.FLOAT_ARRAY,
                              pyferret.STRING_ONEVAL,
                              pyferret.STRING_ONEVAL, ),
                "influences": ( (False, False, False, False, False, False),
                                (False, False, False, False, False, False),
                                (False, False, False, False, False, False),
                                (False, False, False, False, False, False), ),
              }
    return retdict


def ferret_result_limits(efid):
    """
    Abstract axis limits for the shapefile_writeval PyEF
    """
    return ( (1, 1), None, None, None, None, None, )


def ferret_compute(efid, result, resbdf, inputs, inpbdfs):
    """
    Create the shapefile named in inputs[0] using the values array
    given in inputs[1].  The bounding box limits of the X and Y axes
    of inputs[1] are used to create the quadrilaterals.  The values
    in inputs[1] are used as the values associated with each shape
    using the field name given in inputs[2].  Either a common name
    or a WKT description of the map projection for the coordinated
    should be given in inputs[3].  If blank, WGS 84 is used.  If
    successful, fills result (which might as well be a 1x1x1x1 array)
    with zeros.  If a problem occurs, an error will be raised.
    """
    shapefile_name = inputs[0]
    values = inputs[1]
    missing_value = inpbdfs[1]
    field_name = inputs[2].strip()
    if not field_name:
        field_name = "VALUE"
    map_projection = inputs[3]

    # Verify the shapes are as expected
    if (values.shape[3] != 1) or (values.shape[4] != 1) or (values.shape[5] != 1):
        raise ValueError("The T, E, and F axes of VALUE must be undefined or singleton axes")

    # Get the X axis box limits for the quadrilateral coordinates
    x_box_limits = pyferret.get_axis_box_limits(efid, pyferret.ARG2, pyferret.X_AXIS)
    if x_box_limits == None:
        raise ValueError("Unable to determine the X axis box limits")
    lowerxs = x_box_limits[0]
    upperxs = x_box_limits[1]

    # Get the Y axis box limits for the quadrilateral coordinates
    y_box_limits = pyferret.get_axis_box_limits(efid, pyferret.ARG2, pyferret.Y_AXIS)
    if y_box_limits == None:
        raise ValueError("Unable to determine the y axis box limits")
    lowerys = y_box_limits[0]
    upperys = y_box_limits[1]

    # Get the elevation/depth coordinates, or None if they do not exist
    z_coords = pyferret.get_axis_coordinates(efid, pyferret.ARG2, pyferret.Z_AXIS)

    # Create polygons with a single field value
    if z_coords == None:
        sfwriter = shapefile.Writer(shapefile.POLYGON)
    else:
        sfwriter = shapefile.Writer(shapefile.POLYGONZ)
    sfwriter.field(field_name, "N", 20, 7)

    # Write out the shapes and the records
    shape_written = False
    if z_coords == None:
        for j in range(len(lowerys)):
            for i in range(len(lowerxs)):
                if values[i, j, 0, 0, 0, 0] != missing_value:
                    pyferret.fershp.addquadxyvalues(sfwriter,
                                    ( lowerxs[i], lowerys[j] ),
                                    ( lowerxs[i], upperys[j] ),
                                    ( upperxs[i], upperys[j] ),
                                    ( upperxs[i], lowerys[j] ),
                                    None,
                                    [ float(values[i, j, 0, 0, 0, 0]) ])
                    shape_written = True
    else:
        for k in range(len(z_coords)):
            for j in range(len(lowerys)):
                for i in range(len(lowerxs)):
                    if values[i, j, k, 0, 0, 0] != missing_value:
                        pyferret.fershp.addquadxyvalues(sfwriter,
                                        ( lowerxs[i], lowerys[j] ),
                                        ( lowerxs[i], upperys[j] ),
                                        ( upperxs[i], upperys[j] ),
                                        ( upperxs[i], lowerys[j] ),
                                        z_coords[k],
                                        [ float(values[i, j, k, 0, 0, 0]) ])
                        shape_written = True
    if not shape_written:
        raise ValueError("All values are missing values")
    sfwriter.save(shapefile_name)

    # Create the .prj file from the map projection common name or the WKT description
    pyferret.fershp.createprjfile(map_projection, shapefile_name)
    result[:, :, :, :, :, :] = 0


#
# The following is only for testing this module from the command line
#
if __name__ == "__main__":
    # make sure these calls do not generate errors
    info = ferret_init(0)
    del info
    limits = ferret_result_limits(0)
    del limits
    # Testing ferret_compute difficult due to call
    # to get_axis_box_limits and get_axis_coordinates

    print("shapefile_writeval: SUCCESS (limited)")



# To be run in python after importing and starting pyferret
# such as from running "pyferret -python"

from __future__ import print_function

import sys ; sys.ps1 = '' ; sys.ps2 = ''
print()

print(">>> dsetnames = ['./data/ens1.nc', './data/ens2.nc', './data/ens3.nc', './data/ens4.nc']")
dsetnames = ['./data/ens1.nc', './data/ens2.nc', './data/ens3.nc', './data/ens4.nc']

print(">>> onefileagg = pyferret.FerAggDSet(name='onefileagg', dsets=dsetnames[-1:], along='E')")
onefileagg = pyferret.FerAggDSet(name='onefileagg', dsets=dsetnames[-1:], along='E')
print(">>> pyferret.showdata()")
pyferret.showdata()
print(">>> print str(onefileagg)")
print(str(onefileagg))
print(">>> onefileagg.close()")
onefileagg.close()
print(">>> pyferret.showdata()")
pyferret.showdata()
print(">>> print str(onefileagg)")
print(str(onefileagg))

print(">>> twofileagg = pyferret.FerAggDSet(name='twofileagg', dsets=dsetnames[-2:], along='E', hide=True)")
twofileagg = pyferret.FerAggDSet(name='twofileagg', dsets=dsetnames[-2:], along='E', hide=True)
print(">>> pyferret.showdata()")
pyferret.showdata()
print(">>> print str(twofileagg)")
print(str(twofileagg))
print(">>> twofileagg.close()")
twofileagg.close()
print(">>> pyferret.showdata()")
pyferret.showdata()
print(">>> print str(twofileagg)")
print(str(twofileagg))

print(">>> fourfiles = pyferret.FerAggDSet(name='fourfiles', dsets=dsetnames, along='E')")
fourfiles = pyferret.FerAggDSet(name='fourfiles', dsets=dsetnames, along='E')
print(">>> pyferret.showdata()")
pyferret.showdata()
print(">>> fourfiles.show()")
fourfiles.show()
print(">>> print str(fourfiles)")
print(str(fourfiles))
print(">>> fourfiles.getdsetnames()")
fourfiles.getdsetnames()
print(">>> fourfiles.getdsets()")
fourfiles.getdsets()

print(">>> fourfiles.SST.showgrid()")
fourfiles.SST.showgrid()

print(">>> fourfiles.close()")
fourfiles.close()
print(">>> print str(fourfiles)")
print(str(fourfiles))
print(">>> fourfiles.getdsetnames()")
fourfiles.getdsetnames()
print(">>> fourfiles.getdsets()")
fourfiles.getdsets()
print(">>> pyferret.showdata()")
pyferret.showdata()




"""
Module of functions involving great circles
(thus assuming spheroid model of the earth)
with points given in longitudes and latitudes.
"""

from __future__ import print_function

import math
import numpy
import numpy.random

# Equatorial radius of the earth in kilometers
EARTH_ER = 6378.137

# Authalic radius of the earth in kilometers
EARTH_AR = 6371.007

# Meridional radius of the earth in kilometers
EARTH_MR = 6367.449

# Polar radius of the earth in kilometers
EARTH_PR = 6356.752

DEG2RAD = math.pi / 180.0
RAD2DEG = 180.0 / math.pi
KM2MI = 0.6213712
MI2KM = 1.609344


def lonlatdistance(pt1lon, pt1lat, pt2lon, pt2lat):
    """
    Compute the great circle distance between two points
    on a sphere using the haversine formula.

    Arguments:
        pt1lon - longitude(s) of the first point
        pt1lat - latitude(s) of the first point
        pt2lon - longitude(s) of the second point
        pt2lat - latitude(s) of the second point
    Returns:
        The great circle distance(s) in degrees [0.0, 180.0]
    """
    lon1 = numpy.deg2rad(numpy.asarray(pt1lon, dtype=float))
    lat1 = numpy.deg2rad(numpy.asarray(pt1lat, dtype=float))
    lon2 = numpy.deg2rad(numpy.asarray(pt2lon, dtype=float))
    lat2 = numpy.deg2rad(numpy.asarray(pt2lat, dtype=float))
    dellat = numpy.power(numpy.sin(0.5 * (lat2 - lat1)), 2.0)
    dellon = numpy.cos(lat1) * numpy.cos(lat2) * \
             numpy.power(numpy.sin(0.5 * (lon2 - lon1)), 2.0)
    dist = 2.0 * numpy.arcsin(numpy.power(dellon + dellat, 0.5))
    return numpy.rad2deg(dist)


def lonlatintersect(gc1lon1, gc1lat1, gc1lon2, gc1lat2,
                    gc2lon1, gc2lat1, gc2lon2, gc2lat2):
    """
    Compute the intersections of two great circles.  Uses the line of
    intersection between the two planes of the great circles.

    Arguments:
        gc1lon1 - longitude(s) of the first point on the first great circle
        gc1lat1 - latitude(s) of the first point on the first great circle
        gc1lon2 - longitude(s) of the second point on the first great circle
        gc1lat2 - latitude(s) of the second point on the first great circle
        gc2lon1 - longitude(s) of the first point on the second great circle
        gc2lat1 - latitude(s) of the first point on the second great circle
        gc2lon2 - longitude(s) of the second point on the second great circle
        gc2lat2 - latitude(s) of the second point on the second great circle
    Returns:
        ( (pt1lon, pt1lat), (pt2lon, pt2lat) ) - the longitudes and latitudes
                  of the two intersections of the two great circles.  NaN will
                  be returned for both longitudes and latitudes if a great
                  circle is not well-defined, or the two great-circles coincide.
    """
    # Minimum acceptable norm of a cross product
    # arcsin(1.0E-7) = 0.02" or 0.64 m on the Earth
    MIN_NORM = 1.0E-7
    # Convert longitudes and latitudes to points on a unit sphere
    # The "+ 0.0 * ptlonr" is to broadcast gcz if needed
    ptlonr = numpy.deg2rad(numpy.asarray(gc1lon1, dtype=float))
    ptlatr = numpy.deg2rad(numpy.asarray(gc1lat1, dtype=float))
    gcz = numpy.sin(ptlatr) + 0.0 * ptlonr
    coslat = numpy.cos(ptlatr)
    gcy = coslat * numpy.sin(ptlonr)
    gcx = coslat * numpy.cos(ptlonr)
    gc1xyz1 = numpy.array([gcx, gcy, gcz])
    #
    ptlonr = numpy.deg2rad(numpy.asarray(gc1lon2, dtype=float))
    ptlatr = numpy.deg2rad(numpy.asarray(gc1lat2, dtype=float))
    gcz = numpy.sin(ptlatr) + 0.0 * ptlonr
    coslat = numpy.cos(ptlatr)
    gcy = coslat * numpy.sin(ptlonr)
    gcx = coslat * numpy.cos(ptlonr)
    gc1xyz2 = numpy.array([gcx, gcy, gcz])
    #
    ptlonr = numpy.deg2rad(numpy.asarray(gc2lon1, dtype=float))
    ptlatr = numpy.deg2rad(numpy.asarray(gc2lat1, dtype=float))
    gcz = numpy.sin(ptlatr) + 0.0 * ptlonr
    coslat = numpy.cos(ptlatr)
    gcy = coslat * numpy.sin(ptlonr)
    gcx = coslat * numpy.cos(ptlonr)
    gc2xyz1 = numpy.array([gcx, gcy, gcz])
    #
    ptlonr = numpy.deg2rad(numpy.asarray(gc2lon2, dtype=float))
    ptlatr = numpy.deg2rad(numpy.asarray(gc2lat2, dtype=float))
    gcz = numpy.sin(ptlatr) + 0.0 * ptlonr
    coslat = numpy.cos(ptlatr)
    gcy = coslat * numpy.sin(ptlonr)
    gcx = coslat * numpy.cos(ptlonr)
    gc2xyz2 = numpy.array([gcx, gcy, gcz])
    # Get the unit-perpendicular to the plane going through the
    # origin and the two points on each great circle.  If the
    # norm of the cross product is too small, the great circle
    # is not well-defined, so zero it out so NaN is produced.
    gc1pp = numpy.cross(gc1xyz1, gc1xyz2, axis=0)
    norm = (gc1pp[0]**2 + gc1pp[1]**2 + gc1pp[2]**2)**0.5
    if len(norm.shape) == 0:
        if numpy.fabs(norm) < MIN_NORM:
            norm = 0.0
    else:
        norm[ numpy.fabs(norm) < MIN_NORM ] = 0.0
    gc1pp /= norm
    gc2pp = numpy.cross(gc2xyz1, gc2xyz2, axis=0)
    norm = (gc2pp[0]**2 + gc2pp[1]**2 + gc2pp[2]**2)**0.5
    if len(norm.shape) == 0:
        if numpy.fabs(norm) < MIN_NORM:
            norm = 0.0
    else:
        norm[ numpy.fabs(norm) < MIN_NORM ] = 0.0
    gc2pp /= norm
    # The line of intersection of the two planes is perpendicular
    # to the two plane-perpendiculars and goes through the origin.
    # Points of intersection are the points on this line one unit
    # from the origin.  If the norm of the cross product is too
    # small, the two planes are practically indistinguishable from
    # each other (coincide).
    pt1xyz = numpy.cross(gc1pp, gc2pp, axis=0)
    norm = (pt1xyz[0]**2 + pt1xyz[1]**2 + pt1xyz[2]**2)**0.5
    if len(norm.shape) == 0:
        if numpy.fabs(norm) < MIN_NORM:
            norm = 0.0
    else:
        norm[ numpy.fabs(norm) < MIN_NORM ] = 0.0
    pt1xyz /= norm
    pt2xyz = -1.0 * pt1xyz
    # Convert back to longitudes and latitudes
    pt1lats = numpy.rad2deg(numpy.arcsin(pt1xyz[2]))
    pt1lons = numpy.rad2deg(numpy.arctan2(pt1xyz[1], pt1xyz[0]))
    pt2lats = numpy.rad2deg(numpy.arcsin(pt2xyz[2]))
    pt2lons = numpy.rad2deg(numpy.arctan2(pt2xyz[1], pt2xyz[0]))
    return ( (pt1lons, pt1lats), (pt2lons, pt2lats) )


def lonlatfwdpt(origlon, origlat, endlon, endlat, fwdfact):
    """
    Find the longitude and latitude of a point that is a given factor
    times the distance along the great circle from an origination point
    to an ending point.

    Note that the shorter great circle arc from the origination point
    to the ending point is always used.

    If O is the origination point, E is the ending point, and P is
    the point returned from this computation, a factor value of:
        0.5: P bisects the great circle arc between O and E
        2.0: E bisects the great circle arc between O and P
       -1.0: O bisects the great circle arc between P and E

    Arguments:
        origlon - longitude(s) of the origination point
        origlat - latitude(s) of the origination point
        endlon  - longitude(s) of the ending point
        endlat  - latitude(s) of the ending point
        fwdfact - forward distance factor(s)

    Returns:
        (ptlon, ptlat) - longitude and latitude of the computed point(s).
                         NaN will be returned for both the longitude and
                         latitude if the great circle is not well-defined.
    """
    # Minimum acceptable norm of a cross product
    # arcsin(1.0E-7) = 0.02" or 0.64 m on the Earth
    MIN_NORM = 1.0E-7
    # Convert longitudes and latitudes to points on a unit sphere
    # The "+ 0.0 * ptlonr" is to broadcast gcz if needed
    ptlonr = numpy.deg2rad(numpy.asarray(origlon, dtype=float))
    ptlatr = numpy.deg2rad(numpy.asarray(origlat, dtype=float))
    gcz = numpy.sin(ptlatr) + 0.0 * ptlonr
    coslat = numpy.cos(ptlatr)
    gcy = coslat * numpy.sin(ptlonr)
    gcx = coslat * numpy.cos(ptlonr)
    origxyz = numpy.array([gcx, gcy, gcz])
    #
    ptlonr = numpy.deg2rad(numpy.asarray(endlon, dtype=float))
    ptlatr = numpy.deg2rad(numpy.asarray(endlat, dtype=float))
    gcz = numpy.sin(ptlatr) + 0.0 * ptlonr
    coslat = numpy.cos(ptlatr)
    gcy = coslat * numpy.sin(ptlonr)
    gcx = coslat * numpy.cos(ptlonr)
    endxyz = numpy.array([gcx, gcy, gcz])
    # Determine the rotation matrix about the origin that takes
    # origxyz to (1,0,0) (equator and prime meridian) and endxyz
    # to (x,y,0) with y > 0 (equator in eastern hemisphere).
    #
    # The first row of the matrix is origxyz.
    #
    # The third row of the matrix is the normalized cross product
    # of origxyz and endxyz.  (The great circle plane perpendicular.)
    # If the norm of this cross product is too small, the great
    # circle is not well-defined, so zero it out so NaN is produced.
    gcpp = numpy.cross(origxyz, endxyz, axis=0)
    norm = (gcpp[0]**2 + gcpp[1]**2 + gcpp[2]**2)**0.5
    if len(norm.shape) == 0:
        if numpy.fabs(norm) < MIN_NORM:
            norm = 0.0
    else:
        norm[ numpy.fabs(norm) < MIN_NORM ] = 0.0
    gcpp /= norm
    # The second row of the matrix is the cross product of the
    # third row (gcpp) and the first row (origxyz).  This will
    # have norm 1.0 since gcpp and origxyz are perpendicular
    # unit vectors.
    fwdax = numpy.cross(gcpp, origxyz, axis=0)
    # Get the coordinates of the rotated end point.
    endtrx = origxyz[0] * endxyz[0] + origxyz[1] * endxyz[1] + origxyz[2] * endxyz[2]
    endtry = fwdax[0]   * endxyz[0] + fwdax[1]   * endxyz[1] + fwdax[2]   * endxyz[2]
    # Get the angle along the equator of the rotated end point, multiply
    # by the given factor, and convert this new angle back to coordinates.
    fwdang  = numpy.arctan2(endtry, endtrx)
    fwdang *= numpy.asarray(fwdfact, dtype=float)
    fwdtrx  = numpy.cos(fwdang)
    fwdtry  = numpy.sin(fwdang)
    # Rotate the new point back to the original coordinate system
    # The inverse rotation matrix is the transpose of that matrix.
    fwdx = origxyz[0] * fwdtrx + fwdax[0] * fwdtry
    fwdy = origxyz[1] * fwdtrx + fwdax[1] * fwdtry
    fwdz = origxyz[2] * fwdtrx + fwdax[2] * fwdtry
    # Convert the point coordinates into longitudes and latitudes
    ptlat = numpy.rad2deg(numpy.arcsin(fwdz))
    ptlon = numpy.rad2deg(numpy.arctan2(fwdy, fwdx))
    return (ptlon, ptlat)


def equidistscatter(min_lon, min_lat, max_lon, max_lat, min_gcdist, dfactor=5.0):
    """
    Create a roughly equidistant set of points in a specified region.

    This is done by creating a dense "grid" of points, then repeatedly
    randomly selecting a point from that collection and eliminating
    points too close to that selected point.  For the special cases
    where min_lon and max_lon, or min_lat and max_lat, are very close
    relative to min_gcdist, the maximum number of evenly spaced points
    that can be put on the line described is computed and assigned.

    Arguments:
        min_lon - minimum longitude of the region
        min_lat - minimum latitude of the region
        max_lon - maximum longitude of the region
        max_lat - maximum latitude of the region
        min_gcdist - minimum distance, in great circle degrees,
                  between returned points
        dfactor - the number of axis points in the dense "grid"
                  compared to the desired "grid".  Larger value will
                  generally increase the uniformity of the returned
                  points but will also increase the time required
                  for the calculation.
    Returns:
        (pt_lons, pt_lats) - ptlons is an array of longitudes and ptlats
                  is an array of latitudes of (somewhat random) points in
                  the specified region that are roughly equidistant from
                  each other but not closer than min_gcdist to each other.
    """
    lonmin = float(min_lon)
    lonmax = float(max_lon)
    if math.fabs(lonmax - lonmin) > 180.0:
        raise ValueError("Difference between max_lon and min_lon is more than 180.0")
    latmin = float(min_lat)
    if math.fabs(latmin) > 90.0:
        raise ValueError("min_lat is not in [-90.0,90.0]")
    latmax = float(max_lat)
    if math.fabs(latmax) > 90.0:
        raise ValueError("max_lat is not in [-90.0,90.0]")
    mindeg = float(min_gcdist)
    if (mindeg <= 0.0) or (mindeg >= 90.0):
        raise ValueError("min_gcdist is not in (0.0,90.0)")
    dfact = float(dfactor)
    if dfact < 1.0:
        raise ValueError("dfactor is less than one");

    # If lonmin is relatively close to lonmax, directly
    # compute the points.  Distance on a meridian is the
    # difference in latitudes.
    if math.fabs(lonmax - lonmin) < (0.05 * mindeg):
        lon = 0.5 * (lonmax + lonmin)
        dellat = mindeg
        numlats = int( (math.fabs(latmax - latmin) + dellat) / dellat )
        if latmax < latmin:
            dellat *= -1.0
        hdiff = 0.5 * ( (latmax - latmin) - (numlats - 1) * dellat )
        latvals = numpy.linspace(latmin + hdiff, latmax - hdiff, numlats)
        lonvals = numpy.ones((numlats,), dtype=float) * lon
        return (lonvals, latvals)

    # If latmin is relatively close to latmax, directly
    # compute the points.  Distance depends on the latitude
    # as well as the difference in longitudes.
    if math.fabs(latmax - latmin) < (0.05 * mindeg):
        lat = 0.5 * (latmax + latmin)
        numer = math.sin(0.5 * DEG2RAD * mindeg)
        denom = math.cos(lat * DEG2RAD)
        if numer < denom:
            dellon = math.asin(numer / denom) * 2.0 * RAD2DEG
            numlons = int( (math.fabs(lonmax - lonmin) + dellon) / dellon )
        else:
            # everything too close to a pole - just select one point
            dellon = 180.0
            numlons = 1
        if lonmax < lonmin:
            dellon *= -1.0
        hdiff = 0.5 * ( (lonmax - lonmin) - (numlons - 1) * dellon )
        lonvals = numpy.linspace(lonmin + hdiff, lonmax - hdiff, numlons)
        latvals = numpy.ones((numlons,), dtype=float) * lat
        return (lonvals, latvals)

    # Get the number of latitudes for the dense grid
    # Always use latmin and latmax, even if they are too close
    dellat = mindeg / dfact
    numlats = int( (math.fabs(latmax - latmin) + dellat) / dellat )
    if numlats < 2:
         numlats = 2
    latvals = numpy.linspace(latmin, latmax, numlats)

    # Create the dense grid of longitudes and latitudes
    denslons = [ ]
    denslats = [ ]
    numer = math.sin(0.5 * DEG2RAD * mindeg / dfact)
    for lat in latvals:
        # Get the number of longitudes for the dense grid
        # Always use lonmin and lonmax, even if they are too close
        denom = math.cos(lat * DEG2RAD)
        if numer < denom:
            dellon = math.asin(numer / denom) * 2.0 * RAD2DEG
            numlons = int( (math.fabs(lonmax - lonmin) + dellon) / dellon )
            if numlons < 2:
                numlons = 2
        else:
            # too close to a pole
            numlons = 2
        lonvals = numpy.linspace(lonmin, lonmax, numlons)
        # Add each lon,lat pair to the dense grid
        for lon in lonvals:
            denslons.append(lon)
            denslats.append(lat)
    denslons = numpy.asarray(denslons)
    denslats = numpy.asarray(denslats)

    # create a random permutation of the indices to use for the selection order
    availinds = numpy.random.permutation(len(denslats))
    selectinds = [ ]
    while len(availinds) > 0:
        # Get the index of the next available point
        ind = availinds[0]
        selectinds.append(ind)
        # Compute the distance of the available points to the selected point
        gcdists = lonlatdistance(denslons[ind], denslats[ind],
                                 denslons[availinds], denslats[availinds])
        # Remove indices of any available points too close to this point
        availinds = availinds[ gcdists >= mindeg ]
    # sort the selected indices so the longitudes and latitudes have some order
    selectinds = numpy.sort(selectinds)
    # get the selected longitudes and latitudes
    selectlons = denslons[selectinds]
    selectlats = denslats[selectinds]
    # return the selected longitudes and latitudes arrays
    return (selectlons, selectlats)

#
# The following is just for testing "by-hand" and to serve as examples.
#

if __name__ == "__main__":
    # Test lonlatdistance
    tenten = numpy.linspace(0.0,90.0,10)
    # On the equator, distance = delta longitude
    dists = lonlatdistance(0.0, 0.0, tenten, 0.0)
    if not numpy.allclose(dists, tenten):
        raise ValueError("Equatorial distances FAIL; expect: %s; found: %s" % (str(tenten), str(dists)))
    print("Equatorial distance: PASS")
    print()

    # On any meridian, distance = delta latitude
    dists = lonlatdistance(20.0, 0.0, 20.0, tenten)
    if not numpy.allclose(dists, tenten):
        raise ValueError("Meridional distances FAIL; expect: %s; found: %s" % (str(tenten), str(dists)))
    print("Meridional distance: PASS")
    print()

    # Play with some distances between cities (deg W, deg N)
    seattle =  (122.0 + (20.0 / 60.0), 47.0 + (37.0 / 60.0))
    portland = (122.0 + (41.0 / 60.0), 45.0 + (31.0 / 60.0))
    spokane =  (117.0 + (26.0 / 60.0), 47.0 + (40.0 / 60.0))
    austin =   ( 97.0 + (45.0 / 60.0), 30.0 + (15.0 / 60.0))
    houston =  ( 95.0 + (23.0 / 60.0), 29.0 + (46.0 / 60.0))
    dallas =   ( 96.0 + (48.0 / 60.0), 32.0 + (47.0 / 60.0))

    lons = ( seattle[0], portland[0], spokane[0] )
    lons1, lons2 = numpy.meshgrid(lons, lons)
    lats = ( seattle[1], portland[1], spokane[1] )
    lats1, lats2 = numpy.meshgrid(lats, lats)
    dists = lonlatdistance(lons1, lats1, lons2, lats2)
    dists *= DEG2RAD * EARTH_MR * KM2MI
    expected = [ [   0, 146, 228 ],
                 [ 146,   0, 290 ],
                 [ 228, 290,   0 ] ]
    if not numpy.allclose(dists, expected, rtol=0.01):
        raise ValueError("Seattle, Portland, Spokane distance matrix in miles\n" \
                         "    expect: %s\n"
                         "    found:  %s" % (str(expected), str(dists)))
    print("Seattle, Portland, Spokane distance matrix: PASS")
    print()

    lons = ( austin[0], houston[0], dallas[0] )
    lons1, lons2 = numpy.meshgrid(lons, lons)
    lats = ( austin[1], houston[1], dallas[1] )
    lats1, lats2 = numpy.meshgrid(lats, lats)
    dists = lonlatdistance(lons1, lats1, lons2, lats2)
    dists *= DEG2RAD * EARTH_MR * KM2MI
    expected = [ [   0, 145, 184 ],
                 [ 145,   0, 224 ],
                 [ 184, 224,   0 ] ]
    if not numpy.allclose(dists, expected, rtol=0.01):
        raise ValueError("Austin, Houston, Dallas distance matrix in miles\n" \
                         "    expect: %s\n"
                         "    found:  %s" % (str(expected), str(dists)))
    print("Austin, Houston, Dallas distance matrix: PASS")
    print()

    # Test lonlatintersect
    # Intersections of the equator with meridians
    ((pt1lons, pt1lats), (pt2lons, pt2lats)) = \
            lonlatintersect(0.0, 0.0, tenten, 0.0, \
                            0.0, -90.0, tenten, tenten)
    # First of the first great circle and last of the second great circle are not well-defined
    expvalid = numpy.array([ True ] + ([ False ]*8) + [ True ])
    validity = numpy.isnan(pt1lons)
    if not numpy.allclose(validity, expvalid):
        raise ValueError("Validity of pt1lons: expect: %s, found: %s" % \
                          (str(expvalid), str(validity)))
    validity = numpy.isnan(pt1lats)
    if not numpy.allclose(validity, expvalid):
        raise ValueError("Validity of pt1lats: expect: %s, found: %s" % \
                          (str(expvalid), str(validity)))
    validity = numpy.isnan(pt2lons)
    if not numpy.allclose(validity, expvalid):
        raise ValueError("Validity of pt2lons: expect: %s, found: %s" % \
                          (str(expvalid), str(validity)))
    validity = numpy.isnan(pt2lats)
    if not numpy.allclose(validity, expvalid):
        raise ValueError("Validity of pt2lats: expect: %s, found: %s" % \
                          (str(expvalid), str(validity)))
    if not numpy.allclose(pt1lons[1:-1], tenten[1:-1]):
        raise ValueError("Valid pt1lons: expect: %s, found: %s" %\
                          (str(tenten[1:-1]), str(pt1lons[1:-1])))
    if not numpy.allclose(pt1lats[1:-1], 0.0):
        raise ValueError("Valid pt1lats: expect: all zeros, found: %s" %\
                          str(pt1lats[1:-1]))
    if not numpy.allclose(pt2lons[1:-1], tenten[1:-1]-180.0):
        raise ValueError("Valid pt2lons: expect: %s, found %s" %\
                          (str(tenten[1:-1]-180.0), str(pt2lons[1:-1])))
    if not numpy.allclose(pt2lats[1:-1], 0.0):
        raise ValueError("Valid pt2lats: expect: all zeros, found %s" %\
                          str(pt2lats[1:-1]))
    print("Equator/meridian intersections: PASS")
    print()

    ((pt1lons, pt1lats), (pt2lons, pt2lats)) = \
            lonlatintersect( 0.0, 89.99, 180.0, 89.99,
                            90.0, 89.99, -90.0, 89.99)
    # longitudes could actually be anything, but this algorithm gives 45.0 and -135.0
    if (abs(pt1lons -  45.0) > 1.0E-8) or (abs(pt1lats - 90.0) > 1.0E-8) or \
       (abs(pt2lons + 135.0) > 1.0E-8) or (abs(pt2lats + 90.0) > 1.0E-8):
        raise ValueError("Mini north pole cross intersections: expect: %s, found %s" % \
                         (str([45.0, 90.0, 135.0, -90.0]),
                          str([float(pt1lons), float(pt1lats),
                               float(pt2lons), float(pt2lats)])))
    print("Mini north pole cross intersections: PASS")
    print()


    # Test lonlatfwdpt
    lons, lats = lonlatfwdpt(portland[0], portland[1], spokane[0], spokane[1], 0.0)
    if not ( numpy.allclose(lons, portland[0]) and numpy.allclose(lats, portland[1]) ):
        raise ValueError("Zero forward from portland to spokane: expect %s, found %s" % \
                         (str(portland), str((lons, lats))))
    print("Zero forward: PASS")
    print()

    lons, lats = lonlatfwdpt(portland[0], portland[1], spokane[0], spokane[1], 1.0)
    if not ( numpy.allclose(lons, spokane[0]) and numpy.allclose(lats, spokane[1]) ):
        raise ValueError("One forward from portland to spokane: expect %s, found %s" % \
                         (str(spokane), str((lons, lats))))
    print("One forward: PASS")
    print()

    lons, lats = lonlatfwdpt(0.0, 0.0, tenten, 0.0, 3.0)
    expectlons = 3.0 * tenten
    expectlons[ expectlons > 180.0 ] -= 360.0
    # The first great circle is not well-defined
    expvalid = numpy.array([ True ] + ([ False ]*9))
    validity = numpy.isnan(lons)
    if not numpy.allclose(validity, expvalid):
        raise ValueError("Validity of fwd equator lons: expect: %s, found: %s" % \
                          (str(expvalid), str(validity)))
    validity = numpy.isnan(lats)
    if not numpy.allclose(validity, expvalid):
        raise ValueError("Validity of fwd equator lats: expect: %s, found: %s" % \
                          (str(expvalid), str(validity)))
    if not numpy.allclose(lons[1:], expectlons[1:]):
        raise ValueError("Valid fwd equator lons: expect: %s, found: %s" %\
                          (str(expectlons[1:]), str(lons[1:])))
    if not numpy.allclose(lats[1:], 0.0):
        raise ValueError("Valid fwd equator lats: expect: all zeros, found: %s" %\
                          str(lats[1:]))
    print("Fwd equator: PASS")
    print()

    lons, lats = lonlatfwdpt(0.0, -90.0, 0.0, tenten, 2.0)
    # First longitude could be anything, but this algorithm gives 0.0
    expectlats = 90.0 - 2.0 * tenten
    # The last great circle is not well-defined
    expvalid = numpy.array(([ False ]*9) + [ True ])
    validity = numpy.isnan(lons)
    if not numpy.allclose(validity, expvalid):
        raise ValueError("Validity of fwd prime meridian lons: expect: %s, found: %s" % \
                          (str(expvalid), str(validity)))
    validity = numpy.isnan(lats)
    if not numpy.allclose(validity, expvalid):
        raise ValueError("Validity of fwd prime meridian lats: expect: %s, found: %s" % \
                          (str(expvalid), str(validity)))
    # First longitude could be anything so ignore it
    # Others should be either 180 == -180
    poslons = lons[1:-1]
    poslons[(poslons < 0.0)] += 360.0
    if not numpy.allclose(poslons, 180.0):
        raise ValueError("Valid fwd prime meridian lons: expect: all 180.0 or -180.0, found: %s" %\
                          str(poslons))
    if not numpy.allclose(lats[:-1], expectlats[:-1]):
        raise ValueError("Valid fwd prime meridian lats: expect: %s, found: %s" %\
                          (str(expectlats[:-1]), str(lats[:-1])))
    print("Fwd prime meridian: PASS")
    print()

    lons, lats = lonlatfwdpt(0.0, 0.0, 45.0, 45.0, (2.0, 3.0, 4.0, 5.0))
    expectlons = [ 135.0, 180.0, -135.0, -45.0 ]
    expectlats = [  45.0,   0.0,  -45.0, -45.0 ]
    if not numpy.allclose(lons, expectlons):
        raise ValueError("Fwd diagonal lons: expect: %s, found: %s" %\
                          (str(expectlons), str(lons)))
    if not numpy.allclose(lats, expectlats):
        raise ValueError("Fwd diagonal lats: expect: %s, found: %s" %\
                          (str(expectlats), str(lats)))
    print("Fwd diagonal: PASS")
    print()

    # Test equdistscatter
    lons, lats = equidistscatter(0.0, 0.0, 0.0, 0.0, 1.0)
    if (lons.shape != (1,)) or (lons[0] != 0.0) or \
       (lats.shape != (1,)) or (lats[0] != 0.0):
        raise ValueError("Equidistscatter single-point FAIL; \n" \
                         "  expect: ([0.0],[0.0]), \n" \
                         "  found (%s,%s)" % (str(lons), str(lats)))
    print("Equidistscatter single-point PASS")
    print()

    lons, lats = equidistscatter(0.0, 90.0, 90.0, 90.0, 1.0)
    if (lons.shape != (1,)) or (lons[0] != 45.0) or \
       (lats.shape != (1,)) or (lats[0] != 90.0):
        raise ValueError("Equidistscatter pole-point FAIL; \n" \
                         "  expect: ([45.0],[90.0]), \n" \
                         "  found (%s,%s)" % (str(lons), str(lats)))
    print("Equidistscatter pole-point PASS")
    print()

    lons, lats = equidistscatter(0.0, 0.0, 90.0, 0.0, 1.0)
    if not numpy.all( lats == 0.0 ):
        raise ValueError("Equidistscatter equatorial FAIL; \n" \
                         "  expect: all zero latitudes, \n" \
                         "  found %s" % str(lats))
    deltas = lons[1:] - lons[:-1]
    if not numpy.all( deltas >= 1.0 ):
        raise ValueError("Equidistscatter equatorial FAIL; \n" \
                         "  expect: longitudes monotonic increasing by at least 1.0 degrees, \n" \
                         "  found %s" % str(lons))
    if not numpy.all( deltas < 1.0001 ):
        raise ValueError("Equidistscatter equatorial FAIL; \n" \
                         "  expect: longitudes monotonic increasing by less than 1.0001 degrees, \n" \
                         "  found %s" % str(lons))
    print("Equidistscatter equatorial PASS")
    print()

    lons, lats = equidistscatter(0.0, 0.0, 0.0, 90.0, 1.0)
    if not numpy.all( lons == 0.0 ):
        raise ValueError("Equidistscatter meridional FAIL; \n" \
                         "  expect: all zero longitudes, \n" \
                         "  found %s" % str(lons))
    deltas = lats[1:] - lats[:-1]
    if not numpy.all( deltas >= 1.0 ):
        raise ValueError("Equidistscatter meridional FAIL; \n" \
                         "  expect: latitudes monotonic increasing by at least 1.0 degrees, \n" \
                         "  found %s" % str(lats))
    if not numpy.all( deltas < 1.0001 ):
        raise ValueError("Equidistscatter meridional FAIL; \n" \
                         "  expect: latitudes monotonic increasing by less than 1.0001 degrees, \n" \
                         "  found %s" % str(lats))
    print("Equidistscatter meridional PASS")
    print()

    lons, lats = equidistscatter(0.0, 0.0, 90.0, 90.0, 5.0, 15.0)
    nndists = [ ]
    for j in range(len(lons)):
        gcdists = lonlatdistance(lons[j], lats[j], lons, lats)
        gcdists[j] = 180.0
        if not numpy.all( gcdists >= 5.0 ):
            raise ValueError("Equidistscatter region FAIL; \n" \
                             "  expect distances[%d] >= 2.0, \n" \
                             "  found %s" % (j, str(gcdists)))
        nndists.append(gcdists.min())
    nndists = numpy.array(nndists)
    if not numpy.all( nndists < 10.0 ):
        raise ValueError("Equidistscatter region FAIL; \n" \
                         "  expect nearest neighbor distances < 10.0, \n" \
                         "  found %s" % str(nndists))
    print("Nearest neighbor distances: \n" \
          "    min = %f, max = %f, mean = %f, stdev = %f" % \
          (nndists.min(), nndists.max(), nndists.mean(), nndists.std()))

    print("Equidistscatter region PASS")
    print()



#! python
#

'''
Module defining the EOFAnalysis class for performing an 
empirical orthogonal function analysis of space-time data.  
Also defines the InvalidStateError exception that is 
raised by some methods in the EOFAnalysis class.

@author: Karl M. Smith
'''

from __future__ import print_function

import math
import numpy
import numpy.linalg



class InvalidStateError(Exception):
    '''
    Exception raised by methods of the EOFAnalysis class when
    an instance is not in a proper state for that function call.
    For example, results-reporting methods raise this error if
    called prior to calling the analyze method.
    '''
    pass



class EOFAnalysis(object):
    '''
    Class for performing an Empirical Orthogonal Function (EOF) analysis 
    of space-time data.  Create an instance of this class with the desired 
    space-time data array.  Call the analyze method of this instance to 
    perform the analysis.  Result-reporting methods can then be called to 
    obtain the results the of analysis.  The separate initialization, 
    analysis, and result-reporting methods allow for the creation of 
    methods to modify the input data prior to analysis, and methods to 
    modify the results prior to reporting.
    '''


    def __init__(self, spacetimedata):
        '''
        Initializes an EOFAnalysis instance using the given space-time data.

        Arguments:
            spacetimedata - matrix-like object of data where each column 
                    is a time series at a given location.  Thus the shape 
                    of the (row-major) data array should be (NT, NL) where 
                    where NT is the number of time values at each location 
                    and NL is the number of locations.

        Returns:
            an EOFAnalysis instance using the given space-time data.

        Raises:
            TypeError of ValueError - if spacetimedata cannot be made into
                    a (2D) numpy.matrix of numeric values.  See: numpy.matrix
            UserWarning - if there is one one time value in each time 
                    series.
        '''
        # matrix of averages of each time series
        self.__tavg = None
        # F matrix in EOF analysis = (origvals - tavg)
        self.__fmat = None
        # R matrix = F.T * F / NT (matrix multiplication)
        # (matrix of covariance means) - not saved
        # eigenvalues of R ordered from smallest to largest
        self.__eigvals = None
        # eigenvectors in columns of R corresponding to the eigenvalues
        self.__eigvecs = None
        # fraction significance of the corresponding EOFs
        self.__fracsignifs = None
        # minimum fraction ever considered significant; 0.01 == 1%
        self.__minsignif = 0.01
        # copy of the original space-time values as an matrix
        self.__origvals = numpy.matrix(spacetimedata, dtype=numpy.float64);
        if self.__origvals.shape[0] == 1:
            raise UserWarning('Only one time value in each time series')


    def analyze(self):
        '''
        Perform the empirical orthogonal function analysis of the
        space-time data contained in this EOFAnalysis instance.

        Arguments:
            None

        Returns:
            None

        Raises:
            eofanlysis.InvalidStateError - if the instance has not been 
                    properly initialized with space-time data.
            numpy.linlag.LinAlgError - if there is a problem with the 
                    analysis (diagonalizing the matrix of covariance means).
            UserWarning - if there is no variation found in the time series
     		    (no significant eigenvalues)
        '''
        if self.__origvals == None:
            raise InvalidStateError(
                    'instance data has not been properly initialized')
        # Create a matrix with each row containing the mean value
        # of each time-series in the original data.
        self.__tavg = numpy.average(self.__origvals, axis=0)
        # F is the data adjusted so each time series has a zero mean value.
        # self.__tavg is automatically repeated in the following 
        self.__fmat = self.__origvals - self.__tavg
        # R is the matrix of mean covariances over time = F.T * F / NT
        # R has shape (NP, NP)
        # Since fmat is a matrix, this is performing matrix multiplication.
        rmat = self.__fmat.T * self.__fmat / float(self.__fmat.shape[0])
        # Compute the eigenvalues and eigenvactors of the R matrix.
        # The numpy.linalg.eigh function returns eigenvalues from smallest 
        # to largest, so the significant EOFs are the last eigenvectors.
        # The eigenvectors are in the columns of the eigvectors matrix.
        (eigvals, eigvecs) = numpy.linalg.eigh(rmat)
        # Just making sure eigvals is an array and eigvecs is a matrix.
        self.__eigvals = numpy.array(eigvals, copy=False)
        self.__eigvecs = numpy.matrix(eigvecs, copy=False)
        # Compute the fraction significance of each EOF, which is the 
        # fraction each eigenvalue adds to the sum of all the eigenvalues.
        # Because R = R.T, all eigenvalues are non-negative or negligible
        # real values.
        eigvalsum = numpy.sum(self.__eigvals)
        if eigvalsum >= 1.0E-10:
            self.__fracsignifs = self.__eigvals / numpy.sum(self.__eigvals)
        else:
            # No significant eigenvalues == no significant EOFs
            self.__fracsignifs = numpy.zeros(self.__eigvals.shape)
            raise UserWarning(
     		'No variation in time series (no significant eigenvalues)')


    def signiffracs(self):
        '''
        Returns the significances of each of the empirical orthogonal
        functions (EOFs).  These values are the fraction of the total
        variability explained by each of the EOFs.

        Arguments:
            None

        Returns:
            an array of length NL, the number of locations (and thus,
            the maximum number of EOFs possible), giving the significance
            of the corresponding EOF.  Ordered from most significant to
            least significant.

        Raises:
            eofanalysis.InvalidStateError - if the analyze method has not 
                    been called.  See: EOFAnalysis.analyze
        '''
        if self.__fracsignifs == None:
            raise InvalidStateError('analyze method has not been called')
        # Return a copy of the computed significance of each EOF
        # ordered from largest to smallest (reversed).
        return numpy.array(self.__fracsignifs[::-1])


    def minsignif(self):
        '''
        Returns the current minimum fractional significance value used for
        determining which empirical orthogonal functions are significant.

        Arguments:
            None

        Returns:
            the current minimum fractions significance value
        '''
        return self.__minsignif


    def setminsignif(self, minsignif):
        '''
        Resets the minimum fractional significance value used for
        determining which empirical orthogonal functions are significant.

        Arguments:
            minsignif - value in the range [1.0E-6, 1.0 - 1.0E-6].
                    The 1.0E-6 (0.0001%) restriction is to circumvent
                    issues with numerical accuracy in the calculating
                    eigenvalues.

        Returns:
            None

        Raises:
            ValueError - if the value cannot be interpreted as a
                    numerical value, if the value is less than 1.0E-6,
                    or if the value is greater than (1.0 - 1.0E-6).
        '''
        newsignif = float(minsignif)
        if newsignif < 1.0E-6:
            raise ValueError('minsignif is less than 1.0E-6')
        if newsignif > 1.0 - 1.0E-6:
            raise ValueError('minsignif is greater than (1.0 - 1.0E-6)')
        self.__minsignif = newsignif


    def numeofs(self):
        '''
        Returns the number of significant empirical orthogonal functions
        (EOFs).  These are EOFs with a fraction significance not less
        than the current minimum significance value, which by default is
        0.01 (1%).  See: EOFAnalysis.minsignif and EOFAnalysis.setminsignif

        Arguments:
            None

        Returns:
            the number of significant EOFs.

        Raises:
            eofanalysis.InvalidStateError - if the analyze method has not 
                    been called.  See: EOFAnalysis.analyze
        '''
        if self.__fracsignifs == None:
            raise InvalidStateError('analyze method has not been called')
        # The fracsignifs are ordered from smallest to largest.
        # Optimized for the usual case where most EOFs are insignificant.
        for k in range(len(self.__fracsignifs)):
            if self.__fracsignifs[-(k+1)] < self.__minsignif:
                return k
        return len(self.__fracsignifs)


    def eofvec(self, num):
        '''
        Returns an empirical orthogonal function (EOF) as an array of
        location values in units of the original data.  Different EOFs
        are orthogonal to each other.  The square of the norm of an
        EOF is the eigenvalue associated with that EOF.

        Arguments:
            num -   positive integer giving the number of the EOF to
                    return.  A value of one gives the most influential
                    EOF, two gives the second-most influential EOF, etc.

        Returns:
            the requested EOF as a 1-D array of location values
            in units of the original data.

        Raises:
            ValueError - if num is not a positive integer or if num
                    is larger than the number of significant EOFs.
                    See: EOFAnalysis.numeofs
            eofanalysis.InvalidStateError - if the analyze method has 
                    not been called.  See: EOFAnalysis.analyze
        '''
        if self.__eigvals == None:
            raise InvalidStateError('analyze method has not been called')
        eofnum = int(num)
        if eofnum <= 0:
            raise ValueError('num is not a positive integer')
        if eofnum > len(self.__eigvals):
            raise ValueError('num is larger than the number of EOFs')
        # Eigenvalues are order from smallest to largest; thus, the
        # most significant EOF is derived from the last eigenvector.
        eofnum *= -1
        if self.__fracsignifs[eofnum] < self.__minsignif:
            raise ValueError(
                    'num is larger than the number of significant EOFs')
        # Eigenvectors are in the columns of the eigenvector matrix.
        # Multiply by the square root of the eigenvector to convert 
        # to units of the original data.
        eofvector = numpy.array(self.__eigvecs[:,eofnum]).squeeze() \
                    * math.sqrt(self.__eigvals[eofnum])
        return eofvector


    def tafvec(self, num):
        '''
        Returns the time-amplitude function (TAF) associated with an
        empirical orthogonal function (EOF) as a unitless array.  TAFs
        are orthogonal to each other.  The square of the norm of a TAF 
        equals the number of time values in each time series.

        Arguments:
            num -   positive integer giving the number of the TAF to
                    return.  A value of one gives the TAF associated
                    with the most influential EOF, two gives the TAF
                    associated with the second-most influential EOF, etc.

        Returns:
            the requested TAF as a 1-D array of unitless values.  The
            length of the array is NT, the number of time values in each
            time series.

        Raises:
            ValueError - if num is not a positive integer or is larger 
                    than the number of significant EOFs.
                    See: EOFAnalysis.numeofs
            eofanalysis.InvalidStateError - if the analyze method has not 
                    been called.  See: EOFAnalysis.analyze
        '''
        if self.__eigvals == None:
            raise InvalidStateError('analyze method has not been called')
        eofnum = int(num)
        if eofnum <= 0:
            raise ValueError('num is not a positive integer')
        if eofnum > len(self.__eigvals):
            raise ValueError('num is larger than the number of EOFs')
        # Eigenvalues are order from smallest to largest; thus, the
        # most significant EOF is derived from the last eigenvector.
        eofnum *= -1
        if self.__fracsignifs[eofnum] < self.__minsignif:
            raise ValueError(
                    'num is larger than the number of significant EOFs')
        # Time series are in the columns of the F matrix and eigenvectors are
        # in the columns of the eigenvector matrix.  Matrix multiplication
        # is being performed, and then the product is turned into a 1-D array.
        # Divide by the square root of the eigenvalue to  partially normalize
        # the vector.
        tafvector = numpy.array(self.__fmat * self.__eigvecs[:,eofnum]) \
                         .squeeze() / math.sqrt(self.__eigvals[eofnum])
        return tafvector


    def nullvec(self, num):
        '''
        Returns a vector of the null-space in this empirical orthogonal 
        function (EOF) analysis.  These are the eigenvectors corresponding 
        to EOFs with negligible significance.  Null-space vectors have a 
        norm of one and are orthogonal to each other.

        Since null-space vectors are orthogonal to the data, there are no 
        time amplitude functions (TAFs) associated with these vectors.
        (Or, if you prefer, the TAF is an array of zeros.)

        Arguments:
            num -   positive integer giving the number of the null-space 
                    vector to return.  There is no hierarchy for ordering 
                    null-space vectors.  The number of null-space vectors 
                    present is the number of locations less the number of
                    significant EOFs.  See: EOFAnalysis.numeofs

        Returns:
            the requested null-space vector as a 1-D array of location values.

        Raises:
            ValueError - if num is not a positive integer or is larger
                    than the number of null-space vectors.
            eofanalysis.InvalidStateError - if the analyze method has not 
                    been called.  See: EOFAnalysis.analyze
        '''
        if self.__eigvals == None:
            raise InvalidStateError('analyze method has not been called')
        nullnum = int(num)
        if nullnum <= 0:
            raise ValueError('num is not a positive integer')
        if nullnum > len(self.__eigvals):
            raise ValueError('num is larger than the number of locations')
        nullnum -= 1
        if self.__fracsignifs[nullnum] >= self.__minsignif:
            raise ValueError(
                    'num is larger than the number of null-space vectors')
        nullvector = numpy.array(self.__eigvecs[:,nullnum]).squeeze()
        return nullvector


    def datapiece(self, num):
        '''
        Returns a space-time matrix of computed data that is the portion of
        the original data explained by the indicated empirical orthogonal
        function (EOF) and corresponding time-amplitude function (TAF).

        When the argument is zero, a matrix of repeated time-series
        averages is returned.

        When the argument is positive, the value returned will be the 
        same as TAFmat * EOFmat.T, where TAFmat is a column matrix of
        the TAF values and EOFmat.T is a row matrix of the EOF values.
        However, a simplified formula is actually used in this method.  

        Summing the returned matrices for all significant EOFs will produce 
        a space-time matrix with negligible difference from the original data.  
        However, see EOFAnalysis.dataexplaned for a direct method for 
        obtaining these sums.

        Arguments:
            num -   non-negative integer.  If zero, returns the matrix of
                    repeated time-series averages; if positive, specifies
                    the EOF and TAF to use for computing the contribution.
                    A value of one computes the most significant contribution,
                    two gives the second-most significant contribution, etc.

        Returns:
            the computed space-time data contribution as designated by
            the argument.

        Raises:
            ValueError - if num is not a non-negative integer or is larger
                    than the number of EOFs.  See: EOFAnalysis.numeofs
            eofanalysis.InvalidStateError - if the analyze method has not 
                    been called.  See: EOFAnalysis.analyze
        '''
        if self.__eigvals == None:
            raise InvalidStateError('analyze method has not been called')
        eofnum = int(num)
        if eofnum == 0:
            # return a copy of the matrix with repeated time-series averages
            return self.__tavg.repeat(self.__origvals.shape[0], axis = 0)
        if eofnum < 0:
            raise ValueError('num is not a non-negative integer')
        if eofnum > len(self.__eigvals):
            raise ValueError('num is larger than the number of EOFs')
        # Eigenvalues are order from smallest to largest; thus, the
        # most significant EOF is derived from the last eigenvector.
        eofnum *= -1
        # Get the eofnum most significant eigenvectors as a matrix
        # vect shape = (np, 1)
        vect = self.__eigvecs[:,eofnum]
        # Project fmat - shape (nt, np) - into the subspace defined by this
        # eigenvector.  Using matrix multiplication; amat shape (nt, 1)
        amat = self.__fmat * vect
        # Bring this projection back into the complete space.
        # Matrix multiplication; vect.T shape (1, np); datapart shape (nt, np)
        datapart = amat * vect.T
        return datapart


    def dataexplained(self, num):
        '''
        Returns a space-time matrix of computed data that is the portion
        of the original data explained by the indicated number of the
        most-influential empirical orthogonal functions (EOFs) and their
        corresponding time amplitude functions (TAFs).

        This will be the same as the sum of the return values from
        EOFAnalysis.datapiece called with arguments from zero to num.
        However, a simplified formula is actually used in this method.  

        Arguments:
            num -   non-negative integer.  If zero, returns the matrix of
                    repeated time-series averages; if positive, specifies
                    the number of most-influential EOFs and TAFs to use in 
                    creating the computed data.  A value of one uses only 
                    the most influential EOF and TAF, two uses the two 
                    most-influential EOFs and TAFs, etc.

        Returns:
            the computed space-time data that can be directly compared
            to the original location-time data.

        Raises:
            ValueError - if num is not a non-negative integer or is larger 
                    than the number of EOFs.  See: EOFAnalysis.numeofs
            eofanalysis.InvalidStateError - if the analyze method has not 
                    been called.  See: EOFAnalysis.analyze
        '''
        if self.__eigvals == None:
            raise InvalidStateError('analyze method has not been called')
        eofnum = int(num)
        if eofnum == 0:
            # return a copy of the matrix with repeated time-series averages
            return self.__tavg.repeat(self.__origvals.shape[0], axis = 0)
        if eofnum < 0:
            raise ValueError('num is not a positive integer')
        if eofnum > len(self.__eigvals):
            raise ValueError('num is larger than the number of EOFs')
        # Eigenvalues are order from smallest to largest; thus, the
        # most significant EOF is derived from the last eigenvector.
        eofnum *= -1
        # Get the eofnum most significant eigenvectors as a matrix;
        # shape = (np, num)
        vecmat = self.__eigvecs[:,-1:eofnum-1:-1]
        # Project fmat - shape (nt, np) - into the subspace defined by these
        # eigenvectors.  Using matrix multiplication; amat shape = (nt, num)
        amat = self.__fmat * vecmat
        # Bring this projection back into the complete space and add the
        # time averages to make the result comparable to the original data
        # Matrix multiplication; vecmat.T shape (num, np) and expdata shape
        # (nt, np)
        expdata = (amat * vecmat.T) + self.__tavg
        return expdata


#
# The following is just for test "by-hand" and to serve as some examples.
# See test_eofanalysis.py for tests using the unittest module.
#

if __name__ == '__main__':
    import pprint
    formatter = pprint.PrettyPrinter()

    # five years of quarter-monthly data
    months = numpy.arange(0.0, 60.0, 0.25)

    # create time series on an annual cycle
    cosT = numpy.cos(months * numpy.pi / 6.0)
    sinT = numpy.sin(months * numpy.pi / 6.0)

    print()
    print('spacetime = [ cosT, sinT + 1 ]')
    spacetimedata = numpy.matrix([cosT,
                                  sinT + 1.0]).T
    eofanal = EOFAnalysis(spacetimedata)
    eofanal.analyze()
    defminsignif = eofanal.minsignif();
    eofanal.setminsignif(0.1)
    fracsignifs = eofanal.signiffracs()
    print('EOF fractional significances:')
    print(formatter.pformat(fracsignifs))
    numeofs = eofanal.numeofs()
    print('Number of significant EOFs: %d' % numeofs)
    totalcontrib = eofanal.datapiece(0)
    for k in range(1, numeofs+1):
        eofvec = eofanal.eofvec(k)
        sqnorm = numpy.dot(eofvec, eofvec)
        print('EOF %d has norm^2: %#.4f' % (k, sqnorm))
        print(formatter.pformat(eofvec))
        tafvec = eofanal.tafvec(k)
        sqnorm = numpy.dot(tafvec, tafvec)
        print('TAF %d has norm^2: %#.4f' % (k, sqnorm))
        print(formatter.pformat(tafvec))
        tafeof = numpy.outer(tafvec, eofvec) 
        contrib = eofanal.datapiece(k)
        if numpy.allclose(contrib, tafeof):
            print('datapiece(%d) all close to expected values' % k)
        else:
            raise ValueError(
                    'datapiece(%d):\n    expected\n%s\n    found\n%s' % \
                    (k, formatter.pformat(tafeof),
                        formatter.pformat(contrib)))
        totalcontrib += contrib
        expdata = eofanal.dataexplained(k)
        if numpy.allclose(expdata, totalcontrib):
            print('dataexplained(%d) all close to expected values' % k)
        else:
            raise ValueError(
                    'dataexplained(%d):\n    expected\n%s\n    found\n%s' % \
                    (k, formatter.pformat(totalcontrib),
                        formatter.pformat(expdata)))
        datadeltas = numpy.array(numpy.abs(spacetimedata - expdata)).flatten()
        maxdiff = numpy.max(datadeltas)
        rmsdiff = math.sqrt(numpy.average(datadeltas * datadeltas))
        print('Max and RMS diff btwn data explained by %d most' % k)
        print('    influential EOFs and actual data: %#.8f, %#.8f' % \
              (maxdiff, rmsdiff))
    datadeltas = numpy.array(numpy.abs(spacetimedata - totalcontrib)).flatten()
    maxdiff = numpy.max(datadeltas)
    rmsdiff = math.sqrt(numpy.average(datadeltas * datadeltas))
    print('Max and RMS diff btwn sum of all significant')
    print('    data pieces and actual data: %#.8f, %#.8f' % \
          (maxdiff, rmsdiff))

    print()
    print('spacetime = [ cosT * cosT, cosT * sinT + 1, cosT * sinT + 2, sinT * sinT + 3 ]')
    spacetimedata = numpy.matrix([ cosT * cosT,
                                   cosT * sinT + 1.0,
                                   cosT * sinT + 2.0,
                                   sinT * sinT + 3.0 ]).T
    eofanal = EOFAnalysis(spacetimedata)
    eofanal.analyze()
    fracsignifs = eofanal.signiffracs()
    print('EOF fractional significances:')
    print(formatter.pformat(fracsignifs))
    numeofs = eofanal.numeofs()
    print('Number of significant EOFs: %d' % numeofs)
    totalcontrib = eofanal.datapiece(0)
    for k in range(1, numeofs+1):
        eofvec = eofanal.eofvec(k)
        sqnorm = numpy.dot(eofvec, eofvec)
        print('EOF %d has norm^2: %#.4f' % (k, sqnorm))
        print(formatter.pformat(eofvec))
        tafvec = eofanal.tafvec(k)
        sqnorm = numpy.dot(tafvec, tafvec)
        print('TAF %d has norm^2: %#.4f' % (k, sqnorm))
        print(formatter.pformat(tafvec))
        tafeof = numpy.outer(tafvec, eofvec) 
        contrib = eofanal.datapiece(k)
        if numpy.allclose(contrib, tafeof):
            print('datapiece(%d) all close to expected values' % k)
        else:
            raise ValueError(
                    'datapiece(%d):\n    expected\n%s\n    found\n%s' % \
                    (k, formatter.pformat(tafeof),
                        formatter.pformat(contrib)))
        totalcontrib += contrib
        expdata = eofanal.dataexplained(k)
        if numpy.allclose(expdata, totalcontrib):
            print('dataexplained(%d) all close to expected values' % k)
        else:
            raise ValueError(
                    'dataexplained(%d):\n    expected\n%s\n    found\n%s' % \
                    (k, formatter.pformat(totalcontrib),
                        formatter.pformat(expdata)))
        datadeltas = numpy.array(numpy.abs(spacetimedata - expdata)).flatten()
        maxdiff = numpy.max(datadeltas)
        rmsdiff = math.sqrt(numpy.average(datadeltas * datadeltas))
        print('Max and RMS diff btwn data explained by %d most' % k)
        print('    influential EOFs and actual data: %#.8f, %#.8f' % \
              (maxdiff, rmsdiff))
    datadeltas = numpy.array(numpy.abs(spacetimedata - totalcontrib)).flatten()
    maxdiff = numpy.max(datadeltas)
    rmsdiff = math.sqrt(numpy.average(datadeltas * datadeltas))
    print('Max and RMS diff btwn sum of all significant')
    print('    data pieces and actual data: %#.8f, %#.8f' % \
          (maxdiff, rmsdiff))

    fmat = spacetimedata - eofanal.datapiece(0)
    for k in range(1, spacetimedata.shape[1] - numeofs + 1):
        nullvec = eofanal.nullvec(k)
        sqnorm = numpy.dot(nullvec, nullvec)
        print('Null-space vector %d has norm^2: %#.4f' % (k, sqnorm))
        print(formatter.pformat(nullvec))
        tafvec = numpy.array(fmat * numpy.matrix(nullvec).T).squeeze()
        sqnorm = numpy.dot(tafvec, tafvec)
        print('F * NSV %d has norm^2: %#.4f' % (k, sqnorm))
        # print formatter.pformat(tafvec)

    import pickle

    eofpicklestring = pickle.dumps(eofanal)
    print('length of the eofanal pickle string: %d' % len(eofpicklestring))
    neweofanal = pickle.loads(eofpicklestring)
    print('unpickled eofanal.numeofs() = %d' % neweofanal.numeofs())

